""" Function to iteratively fill depressions in a raster surface. """

import numpy as np
cimport numpy as np
cimport cython

#@cython.profile(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef int find_min(double[:] Z, int[:] FLAG):
    """ Return index of smallest item in Z where FLAG == 1 """
    cdef int itr, idx
    cdef double mn
    mn = 1e16
    for itr in range(len(Z)):
        if FLAG[itr] == 1:
            if Z[itr] < mn:
                idx = itr
                mn = Z[itr]
    return idx

#@cython.profile(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.always_allow_keywords(False)
cdef double[:] flatten(double[:,:] A):
    cdef double[:] Aflat
    cdef int i, j, idx
    Aflat = np.empty(A.size)
    for i in range(A.shape[0]):
        for j in range(A.shape[1]):
            idx = i * A.shape[1] + j
            Aflat[idx] = A[i,j]
    return Aflat

#@cython.profile(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def fill_sinks_cy2(double[:,:] Z):
    """ Fill sinks in a DEM following the algorithm of Wang and Liu
    (2006).

        *Z*    :    2d array of elevation or potential data
                    (must not contain NaN!)

    Wang, L. and Liu, H. An efficient method for identifying and filling
    surface depressions in digital elevation models for hydrologic
    analysis and modelling. International Journal of Geographical
    Information Science, 20:2 (2006).
    """

    cdef int[:] OPENROW, OPENCOL, OPENIND, BROW, BCOL, CLOSED
    cdef double[:] ZFLAT, OPENZ, SPILL
    cdef int n, nx, ny, nopen, i, j      # counters and row/column indices
    cdef int bidx, oidx, nidx     # boundary, open, and neighbour indices
    cdef double Zn
    cdef tuple neighbours
    
    # Initialize SPILL, OPEN and CLOSED
    n = Z.shape[0] * Z.shape[1]
    ZFLAT = flatten(Z)
    SPILL = np.zeros(n, dtype=np.double)
    CLOSED = np.zeros(n, dtype=np.int32)
    OPENZ = np.zeros(n, dtype=np.double)
    OPENIND = np.zeros(n, dtype=np.int32)
    BROW = np.zeros(n, dtype=np.int32)
    BCOL = np.zeros(n, dtype=np.int32)

    # Get the boundary cells
    ny, nx = (Z.shape[0], Z.shape[1])
    bidx = 0
    for i in range(ny):
        BROW[bidx] = i
        BCOL[bidx] = 0
        bidx += 1
        BROW[bidx] = i
        BCOL[bidx] = nx-1
        bidx += 1

    for j in range(nx):
        BROW[bidx] = 0
        BCOL[bidx] = j
        bidx += 1
        BROW[bidx] = ny-1
        BCOL[bidx] = j
        bidx += 1

    # Add boundary to OPEN
    nopen = 0
    for i in range(bidx):
        oidx = BROW[i]*nx + BCOL[i]
        OPENZ[oidx] = ZFLAT[oidx]
        OPENIND[oidx] = 1
        nopen += 1
        SPILL[oidx] = ZFLAT[oidx]
    
    # Iteratively fill sinks
    while nopen > 0:

        oidx = find_min(OPENZ, OPENIND)
        OPENIND[oidx] = 0
        nopen -= 1
        CLOSED[oidx] = 1

        neighbours = (oidx-nx-1, oidx-nx, oidx-nx+1,
                      oidx-1,             oidx+1,
                      oidx+nx-1, oidx+nx, oidx+nx+1)

        for nidx in neighbours:

            if 0 <= nidx < n:
                if CLOSED[nidx] == 0 and OPENIND[nidx] == 0:
    
                    Zn = ZFLAT[nidx]
                    SPILL[nidx] = max(Zn, SPILL[oidx])
    
                    OPENZ[nidx] = Zn
                    OPENIND[nidx] = 1
                    nopen += 1

    return np.reshape(SPILL, (ny, nx))



""" Function to iteratively fill depressions in a raster surface. """

import heapq
import numpy as np

def neighbours_of(a):
    """ For a (z,i,j) point `a`, return the neighbouring indices. """
    _, i, j = a
    return ((i-1, j-1), (i, j-1), (i+1, j-1),
            (i-1, j), (i+1, j),
            (i-1, j+1), (i, j+1), (i+1, j+1))

def fill_sinks(Z):
    """ Fill sinks in a DEM following the algorithm of Wang and Liu
    (2006).

        *Z*    :    2d array of elevation or potential data
                    (must not contain NaN!)

    Wang, L. and Liu, H. An efficient method for identifying and filling
    surface depressions in digital elevation models for hydrologic
    analysis and modelling. International Journal of Geographical
    Information Science, 20:2 (2006).
    """

    # Initialize SPILL and CLOSED
    SPILL = Z.copy()
    CLOSED = np.zeros_like(Z)
    OPEN = []
    nopen = 0

    # Get the boundary cells
    ny, nx = (Z.shape[0], Z.shape[1])
    B = [(i, j) for i in range(ny) for j in (0, nx-1)]
    B.extend([(i, j) for i in (0, ny-1) for j in range(1, nx-1)])

    # Get z along the boundary
    for b in B:
        SPILL[b] = Z[b]
        heapq.heappush(OPEN, (Z[b], b[0], b[1]))
        nopen += 1

    while nopen > 0:

        c = heapq.heappop(OPEN)
        nopen -= 1
        CLOSED[c[1:]] = 1

        for n in neighbours_of(c):
            if (0 <= n[0] < ny) and (0 <= n[1] < nx):
                Zn = Z[n]
                if not CLOSED[n] and (Zn, n[0], n[1]) not in list(OPEN):
                    SPILL[n] = max(Zn, SPILL[c[1], c[2]])
                    heapq.heappush(OPEN, (Zn, n[0], n[1]))
                    nopen += 1

    return SPILL



"""
Classes for handling raster data.
"""

from . import grid
from . import aaigrid
from . import raster
from . import flow

from .grid import RegularGrid, WarpedGrid, aairead
from .aaigrid import AAIGrid
from .raster import witch_of_agnesi, peaks, pad, slope, aspect, grad, div
from .raster import normed_vector_field

try:
    from .crfuncs import streamline2d
except ImportError:
    from .streamline import streamline2d

try:
    from . import cfill_sinks as fill_sinks
except ImportError:
    from . import fill_sinks

try:
    from . import flow          # Has a Scipy dependency
except ImportError:
    pass

__all__ = ["grid", "aaigrid", "raster", "flow",
           "RegularGrid", "WarpedGrid", "aairead",
           "AAIGrid",
           "pad", "slope", "aspect", "grad", "div", "normed_vector_field",
           "streamline2d",
           "fill_sinks"]




"""
Flow-routing algorithms


The following license applies to:
    pixel_flow, facet_flow

    Copyright (c) 2009, The MathWorks, Inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or
    without modification, are permitted provided that the following
    conditions are met:

        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.

        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following
          disclaimer in the documentation and/or other materials
          provided with the distribution

        * Neither the name of the The MathWorks, Inc. nor the names
          of its contributors may be used to endorse or promote
          products derived from this software without specific prior
          written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
    USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
    IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
    THE POSSIBILITY OF SUCH DAMAGE.

"""

#import sys
import math
import numpy as np
from . import raster
try:
    from scipy import sparse
except ImportError:
    pass

try:
    from . import crfuncs as rfuncs
except ImportError:
    from . import rfuncs as rfuncs

def facet_flow(e0, e1, e2, d1=1.0, d2=1.0):
    """ Return flow direction and slope for an east-northeast grid
    facet using the method of Tarboton (1997).

                        e2
                        |
                        |
                        | d2
                        |
                        |
            e0 -------- e1
                  d1

    Inspired by blog post by Steven L. Eddins (2007)
    """

    s1 = (e0 - e1) / d1             # Equation (1)
    s2 = (e1 - e2) / d2             # Equation (2)

    r = math.atan2(s2, s1)          # Equation (3)
    s = math.sqrt(s1*s1 + s2*s2)    # Equation (3)

    # Constrain flow direction to lie within or along the edges of the facet
    r, s = r > 0.0 and (r, s) or (0.0, s1)

    diagonal_angle = math.atan2(d2, d1)
    diagonal_distance = math.sqrt(d1*d1 + d2*d2)
    r, s = r < diagonal_angle and \
            (r, s) or (diagonal_angle, (e0 - e2) / diagonal_distance)

    return r, s


def pixel_flow(E, i, j, d1=1.0, d2=1.0):
    """ Downslope flow direction for DEM pixels using method of Tarboton (1997).

    Parameters
    ----------

    E : elevation array

    i : pixel row

    j : pixel column

    d1 : row spacing (default 1.0)

    d2 : column spacing (default 1.0)

    Inspired by blog post by Steven L. Eddins (2007)
    """

    def border_nans(A):
        """ Return indices of nans that touch the border. """
        #nans = np.isnan(A)
        raise Exception("border_nans not implemented")

    m, n = E.shape

    # Compute linear indices at desired locations
    e0_idx = i*(n-0) + j

    # Table 1, page 311
    # Row and column offsets corresponding to e1 and e2 for each entry
    e1_row_offsets = np.array([0, -1, -1,  0,  0,  1,  1,  0])
    e1_col_offsets = np.array([1,  0,  0, -1, -1,  0,  0,  1])

    e2_row_offsets = np.array([-1, -1, -1, -1,  1,  1,  1,  1])
    e2_col_offsets = np.array([ 1,  1, -1, -1, -1, -1,  1,  1])

    # Linear e1 and e2 offsets
    e1_linear_offsets = e1_row_offsets * (n-0) + e1_col_offsets
    e2_linear_offsets = e2_row_offsets * (n-0) + e2_col_offsets

    # Initialize R and S based on the first facet
    Eflat = E.flatten()
    E0 = Eflat[e0_idx]

    E1 = Eflat[e0_idx + e1_linear_offsets[0]]
    E2 = Eflat[e0_idx + e2_linear_offsets[0]]

    R, S = facet_flow(E0, E1, E2, d1, d2)

    # Multipliers ac and af used to convert from east-northeast facet
    # to any other fact (using Eq. 6)
    ac = [0,  1,  1,  2,  2,  3,  3,  4]
    af = [1, -1,  1, -1,  1, -1,  1, -1]

    R = (af[0] * R) + (ac[0] * math.pi / 2.0) if S > 0 else np.nan

    # Compute Rk and Sk corresponding to the k-th facet
    # Where Sk>0 and Sk>S, S=Sk and recompute R(Rk)
    for k in range(1,8):
        E1 = Eflat[e0_idx + e1_linear_offsets[k]]
        E2 = Eflat[e0_idx + e2_linear_offsets[k]]
        Rk, Sk = facet_flow(E0, E1, E2, d1, d2)

        if (Sk > S) and (Sk > 0.0):
            R = (af[k] * Rk) + (ac[k] * math.pi / 2.0)
        S = max(S, Sk)

    return R, S


def dem_flow(D):
    """ Calculate a flow field (aspect and slope) for an array *D*. Return
    aspect (*R*) and slope (*S*).

    Uses the D-infinity method of Tarboton (1997)
    Based on code implemented by Steven L. Eddins (2007)
    """
    Dp = raster.pad(D)
    c = [(i,j) for i in range(1, Dp.shape[0]-1) for j in range(1, Dp.shape[1]-1)]

    ff = map(lambda a: pixel_flow(Dp, a[0], a[1], d1=1.0, d2=1.0), c)
    R = np.array([i[0] for i in ff]).reshape(D.shape)
    S = np.array([i[1] for i in ff]).reshape(D.shape)
    return R, S

def dem_flow2(D):
    """ Calculate aspect and slope using karta.raster functions """
    pi = np.pi
    asp = raster.aspect(D)
    asp[0,:] = 0.5*pi
    asp[-1,:] = 1.5*pi
    asp[:,0] = pi
    asp[:,-1] = 0.0
    asp[0,0] = 0.75*pi
    asp[0,-1] = 0.25*pi
    asp[-1,0] = 1.25*pi
    asp[-1,-1] = 1.75*pi
    slope = raster.slope(D)
    return asp, slope

def prop_dinfty(position, alpha):
    """ Use the D-oo algorithm to return proportion of flow to the center of a
    3x3 grid.

    Parameters:
    -----------

    position : cell location assigned based on the map

                    0   1   2
                    3   *   4
                    5   6   7

    alpha : angle measured counter-clockwise from a vector pointing toward
    position 4
    """
    pi = math.pi

    # beta is the direction to the centre node
    BETA = [1.75, 1.5, 1.25, 0.0, 1.0, 0.25, 0.5, 0.75]
    beta = BETA[position] * pi

    theta = abs(rfuncs.diffrad(alpha, beta))
    if theta <= 0.25*pi and np.isnan(alpha) == False:
        P = (0.25*pi-theta) / (0.25*pi)
    else:
        P = 0.0
    return P


def prop_d8(position, alpha):
    """ Use the D8 algorthm to return proportion of flow to the center of a 3x3
    grid.

    Parameters:
    -----------

    position : cell location assigned based on the map

                    0   1   2
                    3   *   4
                    5   6   7

    alpha : angle measured counter-clockwise from a vector pointing toward
    position 4
    """
    pi = math.pi

    # beta is the direction to the centre node
    BETA = [1.75, 1.5, 1.25, 0.0, 1.0, 0.25, 0.5, 0.75]
    beta = BETA[position] * pi

    theta = abs(rfuncs.diffrad(alpha, beta))
    if theta <= 0.125*pi and np.isnan(alpha) == False:
        P = 1.0
    else:
        P = 0.0
    return P


def upslope_area(F, A, proportion=prop_dinfty):
    """ Calculate upslope area with a sparse matrix formulation.

    Parameters:
    -----------

    F : flow direction array (e.g. from dem_flow())

    A : array providing the area of each grid cell

    proportion : function that defines how flow should be partitioned (e.g. D8,
        D-infinity algorithms).
    """
    m, n = F.shape
    if not hasattr(A, 'shape'):
        A = np.ones_like(F) * A

    # Assemble a flow contribution matrix
    rows = np.zeros(F.size*9)
    cols = np.zeros(F.size*9)
    data = np.zeros(F.size*9)

    cnt = 0
    for i in xrange(m):
        for j in xrange(n):
            i_ = i*n + j
            i0 = i_ - n - 1
            i1 = i_ - n
            i2 = i_ - n + 1
            i3 = i_ - 1
            i4 = i_ + 1
            i5 = i_ + n - 1
            i6 = i_ + n
            i7 = i_ + n + 1

            rows[cnt] = i_
            cols[cnt] = i_
            data[cnt] = A[i,j]
            cnt += 1

            if i > 0:
                if j > 0:
                    rows[cnt] = i_
                    cols[cnt] = i0
                    data[cnt] = -A[i-1,j-1] * proportion(0, F[i-1,j-1])
                    cnt += 1

                rows[cnt] = i_
                cols[cnt] = i1
                data[cnt] = -A[i-1,j]   * proportion(1, F[i-1,j])
                cnt += 1

                if j < n-1:
                    rows[cnt] = i_
                    cols[cnt] = i2
                    data[cnt] = -A[i-1,j+1] * proportion(2, F[i-1,j+1])
                    cnt += 1

            if j > 0:
                rows[cnt] = i_
                cols[cnt] = i3
                data[cnt] = -A[i,j-1]   * proportion(3, F[i,j-1])
                cnt += 1

            if j < n-1:
                rows[cnt] = i_
                cols[cnt] = i4
                data[cnt] = -A[i,j+1]   * proportion(4, F[i,j+1])
                cnt += 1

            if i < m-1:
                if j > 0:
                    rows[cnt] = i_
                    cols[cnt] = i5
                    data[cnt] = -A[i+1,j-1] * proportion(5, F[i+1,j-1])
                    cnt += 1

                rows[cnt] = i_
                cols[cnt] = i6
                data[cnt] = -A[i+1,j]   * proportion(6, F[i+1,j])
                cnt += 1

                if j < n-1:
                    rows[cnt] = i_
                    cols[cnt] = i7
                    data[cnt] = -A[i+1,j+1] * proportion(7, F[i+1,j+1])
                    cnt += 1

    C = sparse.coo_matrix((data, (rows, cols)), shape=(F.size, F.size))

    # Assemble a flow contribution matrix
    #C = sparse.lil_matrix((F.size, F.size))
    #for i in xrange(m):
    #    for j in xrange(n):

    #        i_ = i*n + j
    #        i0 = i_ - n - 1
    #        i1 = i_ - n
    #        i2 = i_ - n + 1
    #        i3 = i_ - 1
    #        i4 = i_ + 1
    #        i5 = i_ + n - 1
    #        i6 = i_ + n
    #        i7 = i_ + n + 1

    #        C[i_,i_] = A[i,j]
    #        if i > 0:
    #            if j > 0:
    #                C[i_,i0] = -A[i-1,j-1] * proportion(0, F[i-1,j-1])
    #            C[i_,i1]     = -A[i-1,j]   * proportion(1, F[i-1,j])
    #            if j < n-1:
    #                C[i_,i2] = -A[i-1,j+1] * proportion(2, F[i-1,j+1])
    #        if j > 0:
    #            C[i_,i3]     = -A[i,j-1]   * proportion(3, F[i,j-1])
    #        if j < n-1:
    #            C[i_,i4]     = -A[i,j+1]   * proportion(4, F[i,j+1])
    #        if i < m-1:
    #            if j > 0:
    #                C[i_,i5] = -A[i+1,j-1] * proportion(5, F[i+1,j-1])
    #            C[i_,i6]     = -A[i+1,j]   * proportion(6, F[i+1,j])
    #            if j < n-1:
    #                C[i_,i7] = -A[i+1,j+1] * proportion(7, F[i+1,j+1])

    # Next, solve the linear problem: C * U_a = \summation{A}
    U = sparse.linalg.spsolve(C.tocsc(), A.flatten() * np.ones(m*n))
    return U.reshape(F.shape)

#def upslope_area_nn(F, A, proportion=prop_dinfty):
#    """ Calculate upslope area with a sparse matrix formulation.
#
#    Parameters:
#    -----------
#
#    F : flow direction array (e.g. from dem_flow())
#
#    A : array providing the area of each grid cell
#
#    proportion : function that defines how flow should be partitioned (e.g. D8,
#        D-infinity algorithms).
#    """
#    m, n = F.shape
#    if not hasattr(A, 'shape'):
#        A = np.ones_like(F) * A
#
#    nnmap = np.nonzero(np.isnan(F) == False)
#    nnsize = len(nnmap[0])
#
#    # Assemble a flow contribution matrix
#    C = sparse.lil_matrix((nnsize+2, nnsize+2))
#    for cell in range(nnsize):
#        i = nnmap[0][cell] + 1
#        j = nnmap[1][cell] + 1
#
#        i_ = i*n + j
#        i0 = i_ - n - 1
#        i1 = i_ - n
#        i2 = i_ - n + 1
#        i3 = i_ - 1
#        i4 = i_ + 1
#        i5 = i_ + n - 1
#        i6 = i_ + n
#        i7 = i_ + n + 1
#
#        C[i_,i_] = A[i,j]
#        if i > 0:
#            if j > 0:
#                C[i_,i0] = -A[i-1,j-1] * proportion(0, F[i-1,j-1])
#            C[i_,i1]     = -A[i-1,j]   * proportion(1, F[i-1,j])
#            if j < n-1:
#                C[i_,i2] = -A[i-1,j+1] * proportion(2, F[i-1,j+1])
#        if j > 0:
#            C[i_,i3]     = -A[i,j-1]   * proportion(3, F[i,j-1])
#        if j < n-1:
#            C[i_,i4]     = -A[i,j+1]   * proportion(4, F[i,j+1])
#        if i < m-1:
#            if j > 0:
#                C[i_,i5] = -A[i+1,j-1] * proportion(5, F[i+1,j-1])
#            C[i_,i6]     = -A[i+1,j]   * proportion(6, F[i+1,j])
#            if j < n-1:
#                C[i_,i7] = -A[i+1,j+1] * proportion(7, F[i+1,j+1])
#
#    # Next, solve the linear problem: C * U_a = \summation{A}
#    Unn = sparse.linalg.spsolve(C.tocsc(), A.flatten())
#
#    # Reconstruct the full array
#    U = np.zeros_like(F)
#    Unndiag = Unn.diagonal()
#    for cell in range(nnsize):
#        i = nnmap[0][cell] + 1
#        j = nnmap[1][cell] + 1
#        U[i,j] = Unnfull[cell]
#
#    return U




""" Low-level functions for computing vector geometry. """

import math
import numpy as np

def intersects(x0, x1, x2, x3, y0, y1, y2, y3):
    """ Determines whether two line segments intersect on a plane. """
    if x1 != x0:
        m0 = float(y1-y0) / float(x1-x0)
    else:
        m0 = 1e37
    if x3 != x2:
        m1 = float(y3-y2) / float(x3-x2)
    else:
        m1 = 1e37
    if m0 == m1:
        return False
    x = float(y1 - y3 + m1*x3 - m0*x1) / float(m1 - m0)
    if (x < max(x0, x1) and x < max(x2, x3) and
        x > min(x0, x1) and x > min(x2, x3)):
        return True
    else:
        return False

def intersections(x0, x1, x2, x3, y0, y1, y2, y3):
    """ Return the point of intersection between two line segments. Returns NaN
    if the lines do not intersect.
    """
    if x1 != x0:
        m0 = float(y1-y0) / float(x1-x0)
    else:
        m0 = 1e37
    if x3 != x2:
        m1 = float(y3-y2) / float(x3-x2)
    else:
        m1 = 1e37
    if m0 == m1:
        return (np.nan, np.nan)
    x = float(m0*x0 - m1*x2 + y2 - y0) / float(m0 - m1)
    if (x < max(x0, x1) and x < max(x2, x3) and
        x > min(x0, x1) and x > min(x2, x3)):
        y = m0 * (x-x0) + y0
        return (x, y)
    else:
        return (np.nan, np.nan)

def distance(pt0, pt1):
    """ Calculate the distance between two points (tuples) """
    d = math.sqrt(sum([abs(a-b)**2 for a, b in zip(pt0, pt1)]))
    return d

def pt_nearest(pt, endpt1, endpt2):
    """ Determines the point on a segment defined by tuples endpt1
    and endpt2 nearest to a point defined by tuple pt.
    Returns the a nested tuple of (point, distance)
    """
    dot2 = lambda v1,v2: float(v1[0]*v2[0] + v1[1]*v2[1])
    proj2 = lambda u,v: [dot2(u,v) / dot2(v,v) * e for e in v]
    dist = lambda u,v: math.sqrt((u[0]-v[0])**2. + (u[1]-v[1])**2.)

    u = (pt[0] - endpt1[0], pt[1] - endpt1[1])
    v = (endpt2[0] - endpt1[0], endpt2[1] - endpt1[1])
    u_on_v = proj2(u,v)
    u_int = (u_on_v[0] + endpt1[0], u_on_v[1] + endpt1[1])
    #dist_u_int = dist(u_int, pt)

    # Determine whether u_int is inside the segment
    # Otherwise return the nearest endpoint
    if endpt1[0] == endpt2[0]:  # Segment is vertical
        if u_int[1] > max(endpt1[1], endpt2[1]):
            return (endpt1[1] > endpt2[1]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        elif u_int[1] < min(endpt1[1], endpt2[1]):
            return (endpt1[1] < endpt2[1]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        else:
            return (u_int, dist(u_int, pt))
    else:
        if u_int[0] > max(endpt1[0], endpt2[0]):
            return (endpt1[0] > endpt2[0]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        elif u_int[0] < min(endpt1[0], endpt2[0]):
            return (endpt1[0] < endpt2[0]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        else:
            return (u_int, dist(u_int, pt))





""" Classes for basic grid types """

import sys
import copy
from math import sqrt
import numpy as np
from . import _aai          # Contains the ascread driver

try:
    from . import _gtiff
    HAS_GDAL = True
except ImportError:
    HAS_GDAL = False

try:
    from .fill_sinks import fill_sinks
except ImportError:
    sys.stderr.write("Compiled fill_sinks not available. Falling back to "
                     "Python version\n")
    from .raster import fill_sinks
try:
    from scipy.interpolate import griddata
    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False

class Grid(object):
    """ Grid baseclass. Don't use this directly except to implement subclasses.
    The primary attributes defined by all Grid-derived classed are _hdr and
    data.
    """

    def __len__(self):
        # Number of bands
        return np.atleast_3d(self.data).shape[2]

    def _equivalent_structure(self, other):
        """ Test whether another object shares an equivalent grid layout. """
        shdr = self.get_hdr()
        ohdr = self.get_hdr()
        for key in shdr:
            if shdr[key] != ohdr[key]:
                return False
        return True

    def _check_hdr(self, hdr):
        """ Check that hdr contains required fields. Intended to be overloaded.
        """
        return

    def get_hdr(self):
        """ Return the header dictionary. """
        return self._hdr

    def set_hdr(self, hdr):
        """ Set the header dictionary. """
        self._check_hdr(hdr)
        self._hdr = hdr
        return

    def clipz(self, bounds):
        """ Clip the z-range in place to bounds = [min, max]. """
        self.data = self.data.clip(bounds[0], bounds[1])
        return

    def max(self):
        """ Return the maximum non-nan in self.data. """
        return self.data[np.isnan(self.data)==False].max()

    def min(self):
        """ Return the minimum non-nan in self.data. """
        return self.data[np.isnan(self.data)==False].min()

    def minmax(self):
        """ Return the minimum and maximum value of data array. """
        A = self.data[np.isnan(self.data)==False]
        return (np.nanmin(self.data), np.nanmax(self.data))

    def copy(self):
        """ Return a copy. """
        return copy.deepcopy(self)


class RegularGrid(Grid):
    """ Regular (structured) grid class. A RegularGrid contains a fixed number
    of rows and columns with a regular spacing and a scalar or vector field
    defined as `Z`.
    """
    def __init__(self, hdr, Z=None):
        """
        Parameters:
        -----------
        hdr : dictionary of header fields, which must contain
            - xllcorner (float)
            - yllcorner (float)
            - nx (int)
            - ny (int)
            - dx (float)
            - dy (float)
            - nbands (int)
        Z : dependent m-dimensional quantity (nrows x ncols x m)
        """
        if ('nbands' not in hdr) and (Z is not None):
            if Z.ndim >= 3:
                hdr['nbands'] = Z.ndim[2]
            else:
                hdr['nbands'] = 1
        self.set_hdr(hdr)

        shape = (self._hdr['ny'], self._hdr['nx'])
        if Z is None:
            Z = np.empty(shape)
        else:
            if Z.shape[:2] != shape:
                raise GridError('Data array `Z` has an invalid shape')
        self.data = Z
        return

    def __add__(self, other):
        if self._equivalent_structure(other):
            return RegularGrid(copy.copy(self.get_hdr()), Z=self.data + other.data)
        else:
            raise NonEquivalentGridError(self, other)

    def __sub__(self, other):
        if self._equivalent_structure(other):
            return RegularGrid(copy.copy(self.get_hdr()), Z=self.data-other.data)
        else:
            raise NonEquivalentGridError(self, other)

    def _check_hdr(self, hdr):
        """ Check that the header contains the required fields. """
        required_fields = ('xllcorner', 'yllcorner', 'nx', 'ny', 'dx', 'dy',
                           'nbands')
        for key in required_fields:
            if key not in hdr:
                raise GridError('Header missing {0}'.format(key))
        assert isinstance(hdr['nx'], int)
        assert isinstance(hdr['ny'], int)

    def coordmesh(self, *args, **kwargs):
        """ Alias for center_coords() """
        return self.center_coords(*args, **kwargs)

    def center_llref(self):
        """ Return the 'lower-left' reference in terms of a center coordinate.
        """
        return (self._hdr['xllcorner'] + 0.5*self._hdr['dx'],
                self._hdr['yllcorner'] + 0.5*self._hdr['dy'])

    def center_coords(self):
        """ Return the cell-center coordinates. """
        xllcenter, yllcenter = self.center_llref()
        xurcenter = xllcenter + self._hdr['dx'] * (self._hdr['nx'] - 1)
        yurcenter = yllcenter + self._hdr['dy'] * (self._hdr['ny'] - 1)
        return np.meshgrid(np.linspace(xllcenter, xurcenter, self._hdr['nx']),
                           np.linspace(yurcenter, yllcenter, self._hdr['ny']))

    def vertex_coords(self):
        """ Return the coordinates of vertices. """
        xllcorner = self._hdr['xllcorner']
        yllcorner = self._hdr['yllcorner']
        xurcorner = xllcorner + self._hdr['nx'] * self._hdr['dx']
        yurcorner = yllcorner + self._hdr['ny'] * self._hdr['dy']
        nx = self._hdr['nx']
        ny = self._hdr['ny']
        return np.meshgrid(np.linspace(xllcorner, xurcorner, nx + 1),
                           np.linspace(yurcorner, yllcorner, ny + 1))

    def get_region(self, reference='center'):
        """ Return the region characteristics as a tuple, relative to either
        cell centers or the grid edges. """
        if reference == 'center':
            x0, y0  = self.center_llref()
            n = 0
        elif reference == 'edge':
            x0, y0  = self._hdr['xllcorner'], self._hdr['yllcorner']
            n = 1
        else:
            raise GridError("`reference` must be 'center' or 'edge'")
        return (x0, x0 + self._hdr['dx'] * (self._hdr['nx'] + n),
                y0, y0 + self._hdr['dy'] * (self._hdr['ny'] + n))

    def resample_griddata(self, dx, dy, method='nearest'):
        """ Resample array in-place to have spacing `dx`, `dy' using
        *scipy.griddata*

        Parameters
        ----------

        dx : cell dimension, float

        dy : cell dimension, float

        method : interpolation method, string ('nearest', 'linear')
        """
        xllcenter, yllcenter = self.center_llref()
        xurcenter = xllcenter + self._hdr['dx'] * self._hdr['nx']
        yurcenter = yllcenter + self._hdr['dy'] * self._hdr['ny']
        nx = int((xurcenter - xllcenter) // dx)
        ny = int((yurcenter - yllcenter) // dy)

        xx, yy = self.coordmesh()
        xxi, yyi = np.meshgrid(np.linspace(xllcenter, xurcenter, nx),
                               np.linspace(yllcenter, yurcenter, ny))
        idata = griddata((xx.flatten(), yy.flatten()), self.data.flatten(),
                         (xxi.flatten(), yyi.flatten()), method=method)
        self.data = idata.reshape(ny, nx)[::-1]

        hdr = self.get_hdr()
        hdr['dx'] = dx
        hdr['dy'] = dy
        hdr['nx'] = nx
        hdr['ny'] = ny
        self.set_hdr(hdr)
        return

    def resample(self, dx, dy, method='nearest'):
        """ Resample array in-place to have spacing `dx`, `dy'.

        Parameters
        ----------

        dx : cell dimension, float

        dy : cell dimension, float

        method : interpolation method, string ('nearest',)
        """
        xllcenter, yllcenter = self.center_llref()
        xurcenter = xllcenter + self._hdr['dx'] * self._hdr['nx']
        yurcenter = yllcenter + self._hdr['dy'] * self._hdr['ny']
        nx = int((xurcenter - xllcenter) // dx)
        ny = int((yurcenter - yllcenter) // dy)

        if method == 'nearest':
            rx, ry = dx / self._hdr['dx'], dy / self._hdr['dy']
            I = np.around(np.arange(ry/2, self.data.shape[0], ry)).astype(int)
            J = np.around(np.arange(rx/2, self.data.shape[1], rx)).astype(int)
            JJ, II = np.meshgrid(J, I)
            self.data = self.data[II, JJ]
        else:
            raise NotImplementedError('method "{0}" not '
                                      'implemented'.format(method))

        hdr = self.get_hdr()
        hdr['dx'] = dx
        hdr['dy'] = dy
        hdr['nx'] = nx
        hdr['ny'] = ny
        self.set_hdr(hdr)
        return

    def resize(self, te):
        """ Resize array to fit within extents given by te. If the new
        dimensions are smaller, the data is clipped. If they are larger,
        nan padding is added.

        *te*        :   tuple of center coordinates in the form
                        (xmin, xmax, ymin, ymax).

        Returns None.
        """
        if self.data is not None:
            xmin1, ymin1 = self.center_llref()
            xmax1 = xmin1 + self._hdr['dx'] * (self._hdr['nx'] - 1)
            ymax1 = ymin1 + self._hdr['dy'] * (self._hdr['ny'] - 1)

            xmin2 = te[0]
            xmax2 = te[1]
            ymin2 = te[2]
            ymax2 = te[3]

            data_a = self.data.copy()
            ny = data_a.shape[0]

            # The left side
            Dx = int(np.floor((xmin2-xmin1) / float(self._hdr['dx'])))
            if Dx < 0:
                data_a = np.hstack([np.nan*np.ones([ny, Dx]), data_a])
            elif Dx > 0:
                data_a = data_a[:,Dx:]
            xllcenter = xmin1 + self._hdr['dx'] * Dx

            # The right side
            Dx = int(np.floor((xmax2-xmax1) / float(self._hdr['dx'])))
            if Dx > 0:
                data_a = np.hstack([data_a, np.nan*np.ones([ny, Dx])])
            elif Dx < 0:
                data_a = data_a[:,:Dx]
            self._hdr['nx'] = data_a.shape[1]

            _, nx = data_a.shape

            # The bottom
            Dy = int(np.ceil((ymin2-ymin1) / float(self._hdr['dy'])))
            if Dy < 0:
                data_a = np.vstack([data_a, np.nan*np.ones([Dy, nx])])
            elif Dy > 0:
                data_a = data_a[Dy:,:]
            yllcenter = ymin1 + self._hdr['dy'] * Dy

            # The top
            Dy = int(np.floor((ymax2-ymax1) / float(self._hdr['dy'])))
            if Dy > 0:
                data_a = np.vstack([np.nan*np.ones([Dy, nx]), data_a])
            elif Dy < 0:
                data_a = data_a[:Dy,:]
            self._hdr['ny'] = data_a.shape[0]

            self.data = data_a
            self._hdr['xllcorner'] = xllcenter - 0.5*self._hdr['dx']
            self._hdr['yllcorner'] = yllcenter - 0.5*self._hdr['dy']

        else:
            raise GridError("no data to resize")
        return

    def get_indices(self, x, y):
        """ Return the column and row indices for the point nearest
        geographical coordinates (x, y). """
        if self.data is None:
            raise GridError('No raster to query')

        x0, y0 = self.center_llref()
        dx = self._hdr['dx']
        dy = self._hdr['dy']
        nx = self._hdr['nx']
        ny = self._hdr['ny']

        xi = np.clip((np.around((np.array(x) - x0) / dx)).astype(int),
                     0, nx-1)
        yi = np.clip(ny - (np.around((np.array(y)-y0) / dy)).astype(int) - 1,
                     0, ny-1)
        return xi, yi

    def sample_nearest(self, x, y):
        """ Return the value nearest to (`x`, `y`). Nearest grid center
        sampling scheme. """
        xi, yi = self.get_indices(x, y)
        nx = self._hdr['nx']
        ny = self._hdr['ny']
        if (np.any(xi < 0) or np.any(xi >= nx) or
            np.any(yi < 0) or np.any(yi >= ny)):
            raise GridError("coordinates ({0}, {1}) are outside grid region "
                            "({2}, {3})".format(xi, yi, nx, ny))
        return self.data[yi, xi]

    def sample(self, x, y, method="nearest"):
        """ Return the values nearest (`x`, `y`), where `x` and `y` may be
        equal length vectors. *method* may be one of `nearest`, `linear`. """
        if method == "nearest":
            return self.sample_nearest(x, y)
        elif method == "linear":
            import scipy.interpolate
            Xd, Yd = self.center_coords()
            return scipy.interpolate.griddata((Xd.flat, Yd.flat), self.data.flat,
                                              (x,y), method="linear")
        else:
            raise ValueError("method \"{0}\" not understood".format(method))

    def get_profile(self, segments, resolution=10.0):
        """ Sample along a line defined as `segments`. Does not interpolate.

        Parameters:
        -----------
        segments : iterable containing (x,y) pairs
        resolution : sample spacing

        Returns:
        --------
        profile : ndarray
        """

        z = []
        p = 0
        for s, f in zip(segments[:-1], segments[1:]):

            x0 = s[0]
            y0 = s[1]
            xf = f[0]
            yf = f[1]

            xlen = xf-x0
            ylen = yf-y0
            d = sqrt((xf - x0)**2 + (yf - y0)**2)

            while p < d:
                fd = p / d
                fx = fd*xlen
                fy = fd*ylen
                xi, yi = self.get_indices(x0+fx, y0+fy)
                z.append(self.data[yi, xi])
                p += resolution
            p -= d

        return np.array(z)

    def fill_sinks(self):
        """ Fill depressions. Use the algorithm of Wang and Liu (2006). """
        return RegularGrid(copy.copy(self._hdr), Z=fill_sinks(self.data))

    def as_structured(self):
        """ Return a copy as a WarpedGrid instance. This is a more general
        grid class that has a larger memory footprint but can represent more
        flexible data layouts. """
        Xv, Yv = self.vertex_coords()
        return WarpedGrid(Xv, Yv, self.data.copy())

    def aaiwrite(self, f, reference='corner', nodata_value=-9999):
        """ Save internal data as an ASCII grid. Based on the ESRI standard,
        only isometric grids (i.e. `hdr['dx'] == hdr['dy']` can be saved,
        otherwise `GridIOError` is thrown.

        Parameters:
        -----------
        f : either a file-like object or a filename
        reference : specify a header reference ("center" | "corner")
        nodata_value : specify how nans should be represented (int or float)
        """
        if self.data is None:
            raise GridError("no data to write!")

        if reference not in ('center', 'corner'):
            raise GridIOError("reference in AAIGrid.tofile() must be 'center' or "
                           "'corner'")

        if self._hdr['dx'] != self._hdr['dy']:
            raise GridIOError("ASCII grids require isometric grid cells")

        if not hasattr(f, 'read'):
            f = open(f, "w")

        try:
            data_a = self.data.copy()
            data_a[np.isnan(data_a)] = nodata_value

            f.write("NCOLS {0}\n".format(self._hdr['nx']))
            f.write("NROWS {0}\n".format(self._hdr['ny']))
            if reference == 'center':
                d = self._hdr['dx']
                f.write("XLLCENTER {0}\n".format(self._hdr['xllcorner']+0.5*d))
                f.write("YLLCENTER {0}\n".format(self._hdr['yllcorner']+0.5*d))
            elif reference == 'corner':
                f.write("XLLCORNER {0}\n".format(self._hdr['xllcorner']))
                f.write("YLLCORNER {0}\n".format(self._hdr['yllcorner']))
            f.write("CELLSIZE {0}\n".format(self._hdr['dx']))
            f.write("NODATA_VALUE {0}\n".format(nodata_value))
            f.writelines([str(row).replace(',','')[1:-1] +
                            '\n' for row in data_a.tolist()])
        except Exception as e:
            raise e
        finally:
            f.close()
        return

class WarpedGrid(Grid):
    """ Warped Grid class. A WarpedGrid contains a fixed number of rows
    and columns and a scalar or vector field defined on the z-axis. Grid
    spacing is not necessarily regular.
    """

    def __init__(self, hdr=None, X=None, Y=None, Z=None):
        """
        Parameters:
        -----------
        X : first-dimension coordinates of grid nodes
        Y : second-dimension coordinates of grid nodes
        Z : dependent m-dimensional quantity (nrows x ncols x m)
        hdr : (optional) dictionary of header fields, which must contain
            - xllcenter (float)
            - yllcenter (float)
            - nbands (int)
        """

        if True in (a is None for a in (X,Y,Z)):
            raise GridError('Either all or none of (`X`, `Y`, `Z`) must be '
                            'provided')

        if not (X.shape == Y.shape == Z.shape[:2]):
            raise GridError('All of (`X`, `Y`, `Z`) must share the same '
                            'two-dimensional size')

        if hdr is None:
            hdr = {'xllcorner' : X[-1,0], 'yllcorner' : Y[-1,0]}
            if Z.ndim == 2:
                hdr['nbands'] = 1
            elif Z.ndim == 3:
                hdr['nbands'] = Z.shape[2]
            else:
                raise GridError('`Z` must be of dimension 2 or 3')
            X -= X[-1,0]
            Y -= Y[-1,0]
        self.X = X
        self.Y = Y
        self.data = Z
        self.set_hdr(hdr)
        return

    def __add__(self, other):
        if self._equivalent_structure(other):
            return WarpedGrid(hdr=copy.copy(self.get_hdr()), X=self.X.copy(),
                                  Y=self.Y.copy(), Z=self.data + other.data)
        else:
            raise NonEquivalentGridError(self, other)

    def __sub__(self, other):
        if self._equivalent_structure(other):
            return WarpedGrid(hdr=copy.copy(self.get_hdr()), X=self.X.copy(),
                                  Y=self.Y.copy(), Z=self.data-other.data)
        else:
            raise NonEquivalentGridError(self, other)

    def _check_hdr(self, hdr):
        """ Check that the header contains the required fields. """
        for key in ('xllcorner', 'yllcorner', 'nbands'):
            if key not in hdr:
                raise GridError('Header missing {0}'.format(key))

    def rotate(self, deg, origin=(0.0, 0.0)):
        """ Rotate grid by *deg* degrees counter-clockwise around *origin*. """
        raise NotImplementedError

    def resample(self, X, Y):
        """ Resample internal grid to the points defined by `X`, `Y`. """
        raise NotImplementedError

    def fill_sinks(self):
        """ Fill depressions. Use the algorithm of Wang and Liu (2006). """
        return WarpedGrid(copy.copy(self._hdr),
                              X=self.X.copy(), Y=self.Y.copy(),
                              Z=fill_sinks(self.data))


class GridError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message


class GridIOError(GridError):
    def __init__(self, message=''):
        self.message = message


class NonEquivalentGridError(GridError):
    def __init__(self, A, B, message=''):
        if len(message) == 0:
            self.message = ("{0} and {1} do not share equivalent "
                            "grid layouts".format(A, B))
        else:
            self.message = message


def dummy_hdr(arr):
    if len(arr.shape) > 2:
        nbands = arr.shape[2]
    else:
        nbands = 1
    hdr = {'xllcorner': 0.0, 'yllcorner': 0.0,
           'nx': arr.shape[1], 'ny': arr.shape[0],
           'dx': 1.0, 'dy': 1.0,
           'nbands': nbands}
    return hdr

def aairead(fnm):
    """ Convenience function to open a ESRI ASCII grid and return a RegularGrid
    instance.
    """
    Z, aschdr = _aai.aairead(fnm)
    hdr = {'xllcorner'  : aschdr['xllcorner'],
           'yllcorner'  : aschdr['yllcorner'],
           'nx'         : aschdr['ncols'],
           'ny'         : aschdr['nrows'],
           'dx'         : aschdr['cellsize'],
           'dy'         : aschdr['cellsize'],
           'nbands'     : 1}
    Z[Z==aschdr['nodata_value']] = np.nan
    return RegularGrid(hdr, Z=Z)

def gtiffread(fnm):
    """ Convenience function to open a GeoTIFF and return a RegularGrid
    instance.
    """
    if not HAS_GDAL:
        raise NotImplementedError("Right now, loading GeoTiffs requires GDAL.")
    arr, hdr = _gtiff.read(fnm)
    return RegularGrid(hdr, Z=arr.transpose((1,2,0)).squeeze())




""" IO interface to GeoTiffs using GDAL. """

import gdal
import gdalconst as gc
import numpy as np

gdal.UseExceptions()

def read(fnm):
    """ Read a GeoTiff file and return a numpy array and a dictionary of header
    information. """
    hdr = dict()
    dataset = gdal.Open(fnm, gc.GA_ReadOnly)

    try:
        hdr["nbands"] = dataset.RasterCount
        hdr["nx"] = dataset.RasterXSize
        hdr["ny"] = dataset.RasterYSize
        arr = np.empty((hdr["nbands"], hdr["ny"], hdr["nx"]))

        transform = dataset.GetGeoTransform()
        if transform is not None:
            hdr["dx"] = transform[1]
            hdr["dy"] = -transform[5]
            hdr["xllcorner"] = transform[0]
            hdr["yllcorner"] = transform[3] - hdr["ny"] * hdr["dy"]
        else:
            raise AttributeError("No GeoTransform in GDAL Dataset")

        for i in xrange(1, hdr["nbands"]+1):
            band = dataset.GetRasterBand(i)
            arr[i-1,:,:] = band.ReadAsArray()

        projstr = dataset.GetProjection()
        hdr["crs"] = projstr

    finally:
        dataset = None

    return arr, hdr




""" Cython versions of low-level functions for computing vector geometry. """

import math
import numpy as np
cimport numpy as np

cdef inline double dbl_max(double a, double b): return a if a >= b else b
cdef inline double dbl_min(double a, double b): return a if a <= b else b

cpdef int intersects(double x0, double x1, double x2, double x3,
                     double y0, double y1, double y2, double y3):
    """ Determines whether two line segments intersect on a plane. The line
    segments are specified as ((x0,y0),(x1,y1)), etc."""
    cdef double m0, m1
    cdef double x

    if x1 != x0:
        m0 = (y1-y0) / (x1-x0)
    else:
        m0 = 1e37
    if x3 != x2:
        m1 = (y3-y2) / (x3-x2)
    else:
        m1 = 1e37
    if m0 == m1:
        return False
    x = (y1 - y3 + m1*x3 - m0*x1) / (m1 - m0)
    if (x < dbl_max(x0, x1) and x < dbl_max(x2, x3) and
        x > dbl_min(x0, x1) and x > dbl_min(x2, x3)):
        return True
    else:
        return False

def intersections(double x0, double x1, double x2, double x3,
                 double y0, double y1, double y2, double y3):
    """ Return the point of intersection between two line segments. Returns NaN
    if the lines do not intersect.
    """
    cdef double m0, m1
    cdef double x, y

    if x1!=x0:
        m0 = (y1-y0) / (x1-x0)
    else:
        m0 = 1e16
    if x3!=x2:
        m1 = (y3-y2) / (x3-x2)
    else:
        m1 = 1e16
    if m0 == m1:
        return (np.nan, np.nan)
    x = (m0*x0 - m1*x2 + y2 - y0) / (m0 - m1)
    if (x < dbl_max(x0, x1) and x < dbl_max(x2, x3) and
        x > dbl_min(x0, x1) and x > dbl_min(x2, x3)):
        y = m0 * (x-x0) + y0
        return (x, y)
    else:
        return (np.nan, np.nan)

cdef double cdot2(tuple v1, tuple v2):
    return v1[0] * v2[0] + v1[1] * v2[1]

cdef tuple cproj2(tuple u, tuple v):
    cdef double uv, vv
    uv = cdot2(u,v)
    vv = cdot2(v,v)
    return (uv / vv * v[0], uv / vv * v[1])

def distance(tuple pt0, tuple pt1):
    """ Calculate the distance between two points (tuples) """
    cdef float dsq
    dsq = 0.0
    for i in range(min(len(pt0), len(pt1))):
        dsq += abs(pt0[i] - pt1[i])**2
    return math.sqrt(dsq)

def pt_nearest(tuple pt, tuple endpt1, tuple endpt2):
    """ Determines the point on a segment defined by tuples endpt1
    and endpt2 nearest to a point defined by tuple pt.
    Returns the a nested tuple of (point, distance)
    """
    
    cdef tuple u, v, u_int
    cdef tuple u_on_v
    
    dist = lambda u,v: math.sqrt((u[0]-v[0])**2. + (u[1]-v[1])**2.)

    u = (pt[0] - endpt1[0], pt[1] - endpt1[1])
    v = (endpt2[0] - endpt1[0], endpt2[1] - endpt1[1])
    u_on_v = cproj2(u,v)
    u_int = (u_on_v[0] + endpt1[0], u_on_v[1] + endpt1[1])

    # Determine whether u_int is inside the segment
    # Otherwise return the nearest endpoint
    if endpt1[0] == endpt2[0]:  # Segment is vertical
        if u_int[1] > max(endpt1[1], endpt2[1]):
            return (endpt1[1] > endpt2[1]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        elif u_int[1] < min(endpt1[1], endpt2[1]):
            return (endpt1[1] < endpt2[1]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        else:
            return (u_int, dist(u_int, pt))
    else:
        if u_int[0] > max(endpt1[0], endpt2[0]):
            return (endpt1[0] > endpt2[0]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        elif u_int[0] < min(endpt1[0], endpt2[0]):
            return (endpt1[0] < endpt2[0]
                    and (endpt1, dist(endpt1, pt))
                    or (endpt2, dist(endpt2, pt)))
        else:
            return (u_int, dist(u_int, pt))





""" This module includes the `GeoJSON` class and convenience functions for
dealing with GeoJSON data. The `GeoJSON` class uses the builtin json module and
enforces GeoJSON standards. """

import sys
import copy
import json
from collections import namedtuple
import itertools
from numbers import Number
from math import isnan
import traceback


Point = namedtuple('Point', ['coordinates', 'crs'])
MultiPoint = namedtuple('MultiPoint', ['coordinates', 'crs'])
LineString = namedtuple('LineString', ['coordinates', 'crs'])
MultiLineString = namedtuple('MultiLineString', ['coordinates', 'crs'])
Polygon = namedtuple('Polygon', ['coordinates', 'crs'])
MultiPolygon = namedtuple('MultiPolygon', ['coordinates', 'crs'])
GeometryCollection = namedtuple('GeometryCollection', ['geometries'])


class ExtendedJSONEncoder(json.JSONEncoder):
    """ Numpy-specific numbers prior to 1.9 don't inherit from Python numeric
    ABCs. This class is a hack to coerce numpy values into Python types for
    JSON serialization. """

    def default(self, o):
        try:
            if o.dtype in ("int8", "int16", "int32", "int64"):
                return int(o)
            elif o.dtype in ("float16", "float32", "float64", "float128"):
                return float(o)
            elif o.dtype in ("complex64", "complex128", "complex256"):
                return complex(o)
            else:
                raise TypeError("not a recognized type")
        except (AttributeError, TypeError):
            return json.JSONEncoder.default(self, o)


class GeoJSONWriter(object):
    """ Class for converting guppy objects to GeoJSON strings. Multipoint-based
    opbjects are written as 'Features'.

    CRS defaults to be named. A linked CRS can be used by passing
    `linkedcrs=True`.

    Notes:
    ------
    Does not handle 'FeatureCollection' types (see printFeatureCollection()
    function).
    """
    def __init__(self, gpobj, linkedcrs=False, **kwargs):
        type_equiv = {'Point'       : 'Point',
                      'Multipoint'  : 'MultiPoint',
                      'Line'        : 'LineString',
                      'Polygon'     : 'Polygon'}

        if gpobj._geotype in type_equiv:
            self.typestr = type_equiv[gpobj._geotype]
        else:
            raise TypeError('Input object not a recognized geometry')
        crs = gpobj._crs
        bbox = kwargs.get('bbox', None)

        self.gpobj = gpobj
        self.supobj = {}
        if crs is not None:
            self.add_crs(crs, linkedcrs=linkedcrs)
        if self.typestr != 'Point':
            self.supobj['type'] = 'Feature'
            self.add_bbox(bbox)
            self.add_geometry()
        else:
            self.supobj['type'] = 'Point'
            self.add_coordinates()
        return

    def add_crs(self, crs, linkedcrs, **kw):
        """ Tag the JSON object with coordinate reference system metadata using
        a CRS instance, which is expected the have the interface of crs.CRS.

        If True, `linkedcrs` indicates a web-referenced linked CRS. If False,
        the CRS is taked to be named.

        In the case of a linked CRS, the link address and type can be specified
        using the `href` and `linktype` keyword arguments. If not supplied,
        then it is assumed that crs.urn contains the *href* and crs.type
        contains the *type*.

        For more details, see the GeoJSON specification at:
        http://geojson.org/geojson-spec.html#coordinate-reference-system-objects
        """
        if linkedcrs:
            href = kw.get('href', crs.urn)
            linktype = kw.get('linktype', crs.type)
            self.supobj['crs'] = {'type': 'link',
                                  'properties': {'href': href, 'type': linktype}}

        else:
            self.supobj['crs'] = {'type': 'name',
                                  'properties': {'name': crs.urn}}
        return

    def add_bbox(self, bbox=None):
        """ Tag the JSON object with a bounding box. If *bbox* is None
        (default), the bounds will be calculated from the data. Otherwise,
        *bbox* should be in the form (xmin, xmax, ....) for all dimensions.
        """
        if bbox is None:
            if hasattr(self.gpobj, 'get_extents'):
                bbox = self.gpobj.get_extents()
            else:
                raise AttributeError("Type {0} has no 'get_extents' "
                                     "method".format(type(self.gpobj)))
        self.supobj['bbox'] = bbox
        return

    def add_geometry(self):
        """ Add geometry subobject. This should be called for writing
        Multipoint-derived classes. """
        self.supobj['geometry'] = {'type' : self.typestr}
        self.add_coordinates(self.supobj['geometry'])
        self.add_id()
        self.add_properties()
        return

    def add_coordinates(self, target=None):
        """ Add coordinates. Behaviour depends on the type of the geometry
        object. Target must be None (default) or a dictionary member of the
        JSON object. """
        if target is None:
            target = self.supobj

        if self.gpobj._geotype == 'Polygon':
            target['coordinates'] = [list_rec(self.gpobj.get_vertices())]
            if hasattr(self.gpobj, "subs"):
                for poly in self.gpobj.subs:
                    target['coordinates'].append(list_rec(poly.get_vertices()))
        elif hasattr(self.gpobj, 'get_vertices'):
            target['coordinates'] = list_rec(self.gpobj.get_vertices())
        elif hasattr(self.gpobj, 'get_vertex'):
            target['coordinates'] = list_rec(self.gpobj.get_vertex())
        else:
            raise AttributeError('Geometry object has no vertex method')
        return

    def add_properties(self, target=None):
        """ Add the data from the geometry object's data dictionary. """
        if target is None:
            target = self.supobj
        target['properties'] = {}
        if (False in (a is None for a in self.gpobj.data)):
            if hasattr(self.gpobj.data, 'keys'):
                data = self.gpobj.data
            else:
                data = {'point_data': self.gpobj.data}
            for key in data:
                nan2none = lambda a: None if isinstance(a, Number) and isnan(a) else a
                target['properties'][key] = list(map(nan2none, data.getfield(key)))
        return

    def add_id(self, target=None):
        """ Add an index field. """
        if target is None:
            target = self.supobj
        target['id'] = list(range(len(self.gpobj.get_vertices())))
        return

    def print_json(self):
        """ Print GeoJSON representation. """
        return json.dumps(self.supobj, indent=2, cls=ExtendedJSONEncoder)

    def write_json(self, fout):
        """ Dump internal dict-object to JSON using the builtin `json` module.
        """
        json.dump(self.supobj, fout, indent=2, cls=ExtendedJSONEncoder)
        return


class GeoJSONReader(object):
    """ Class for reading general GeoJSON strings and converting to appropriate
    guppy types. Initialize with a file- or StreamIO-like object, and then use
    the pull* methods to extract geometry types. The equivalency table looks
    like:

    -----------------------------------------------------
    | Guppy class       | GeoJSON geometry object       |
    --------------------+--------------------------------
    | Point             | Point                         |
    |                   |                               |
    | Multipoint        | MultiPoint                    |
    |                   |                               |
    | Line              | Linestring, MultiLineString   |
    |                   |                               |
    | Polygon           | Polygon, MultiPolygon         |
    |                   |                               |
    -----------------------------------------------------
     """

    def __init__(self, finput):
        """ Create a reader-object for a GeoJSON-containing file or StreamIO
        object. Use as::

            with open(`fnm`, 'r') as f:
                reader = GeoJSONReader(f)
        """
        if not hasattr(finput, 'read'):
            with open(finput) as f:
                self.jsondict = json.load(f)
        else:
            self.jsondict = json.load(finput)
        self.crs = self.getcrs()
        return

    def _walk(self, dic, geotype):
        """ Find all instances of `key` using recursion. """
        if 'type' in dic and geotype == dic['type']:
            yield dic
        for k in dic:
            if hasattr(dic[k], 'keys'):
                for val in self._walk(dic[k], geotype):
                    yield val
            elif k == 'features':
                for feature in dic[k]:
                    for val in self._walk(feature, geotype):
                        yield val
        return

    def _contains(self, dic, key):
        """ Return whether *key* exists within the JSON hierarchy *dic*. """
        if key in dic:
            return True
        else:
            for val in dic.values():
                if self._contains(val, key):
                    return True
        return False

    def _getproperties(self):
        """ Read feature properties and return a two dictionaries:
            - one representing singleton values ("properties")
            - one representing per-vertex values ("data")
        """
        raise NotImplementedError
        return {}, {}

    def getcrs(self):
        """ Look in the top level for a `crs` field. Return a tuple that is
        either (name, _), (href, type), or (_, _) for a named, linked, or
        undefined CRS, respectively. """
        crs = self.jsondict.get("crs", None)
        if crs is not None:
            if "name" in crs["properties"]:
                res = (crs["properties"]["name"], None)
            else:
                res = (crs["properties"]["href"], crs["properties"]["type"])
        else:
            res = (None, None)
        return res

    def pull_features(self):
        """ Find and return all Feature objects from a FeatureCollection """
        jsonfeats = [obj for obj in self._walk(self.jsondict, 'Feature')]

        featuretype = lambda a: a.get('geometry', {}).get('type', None)
        ispoint = lambda a: featuretype(a) == 'Point'
        ismultipoint = lambda a: featuretype(a) == 'MultiPoint'
        isline = lambda a: featuretype(a) in ('LineString', 'MultiLineString')
        ispolygon = lambda a: featuretype(a) in ('Polygon', 'MultiPolygon')

        # This is an idiotic algorithm - pre-mapping the types isn't saving
        # time because each feature needs to be iterated through anyway
        points = filter(ispoint, jsonfeats)
        multipoints = filter(ismultipoint, jsonfeats)
        lines = filter(isline, jsonfeats)
        polygons = filter(ispolygon, jsonfeats)

        features = []
        for feat in points:
            features.append((self.pull_points(feat)[0], feat['properties'],
                             feat.get('id', None)))
        for feat in multipoints:
            features.append((self.pull_multipoints(feat)[0], feat['properties'],
                             feat.get('id', None)))
        for feat in lines:
            features.append((self.pull_lines(feat)[0], feat['properties'],
                             feat.get('id', None)))
        for feat in polygons:
            features.append((self.pull_polygons(feat)[0], feat['properties'],
                             feat.get('id', None)))

        return features

    def pull_points(self, dic=None):
        """ Return a list of all geometries that can be coerced into a single
        Point. """
        if dic is None:
            dic = self.jsondict
        jsonpoints = self._walk(dic, 'Point')
        points = []
        for point in jsonpoints:
            points.append(Point(point['coordinates'], self.crs))
        return points

    def pull_multipoints(self, dic=None):
        """ Return a list of all geometries that can be coerced into a single
        Multipoint. """
        if dic is None:
            dic = self.jsondict
        jsonmultipoints = self._walk(dic, 'MultiPoint')
        multipoints = []
        for jsonmultipoint in jsonmultipoints:
            multipoints.append(MultiPoint(jsonmultipoint['coordinates'], self.crs))
        return multipoints

    def pull_lines(self, dic=None):
        """ Return a list of all geometries that can be coerced into a Line.
        """
        if dic is None:
            dic = self.jsondict
        jsonlines = self._walk(dic, 'LineString')
        jsonmultilines = self._walk(dic, 'MultiLineString')
        lines = []
        for jsonline in jsonlines:
            lines.append(LineString(jsonline['coordinates'], self.crs))
        for jsonmultiline in jsonmultilines:
            for vertices in jsonmultiline['coordinates']:
                lines.append(MultiLineString(vertices, self.crs))
        return lines

    def pull_polygons(self, dic=None):
        """ Return a list of all geometries that can be coerced into a Polygon.
        """
        if dic is None:
            dic = self.jsondict
        jsonpolygons = self._walk(dic, 'Polygon')
        jsonmultipolygons = self._walk(dic, 'MultiPolygon')
        polygons = []
        for jsonpolygon in jsonpolygons:
            polygons.append(Polygon(jsonpolygon['coordinates'], self.crs))
        for jsonmultipolygon in jsonmultipolygons:
            polygons.append(Polygon(jsonmultipolygon['coordinates'], self.crs))
        return polygons

    def iter_geometries(self):
        """ Return an iterator through all geometries. """
        itergeo = itertools.chain(self.pull_points(),
                                  self.pull_multipoints(),
                                  self.pull_lines(),
                                  self.pull_polygons())
        return itergeo


def list_rec(A):
    """ Recursively convert nested iterables to nested lists """
    if hasattr(A, '__iter__'):
        return [list_rec(el) for el in A]
    else:
        return A

def printGeometryCollection(gpobj_list, **kwargs):
    """ Given an iterable that returns guppy objects, construct a GeoJSON
    FeatureCollection string.
    """
    geometrylist = []
    for gp in gpobj_list:
        writer = GeoJSONWriter(gp)
        geometrylist.append(copy.copy(writer.supobj))
    geometry_coll = {'type' : 'GeometryCollection',
                    'geometries' : geometrylist}
    kwargs.setdefault("indent", 2)
    return json.dumps(geometry_coll, **kwargs)

def geojson2csv(fin, fout):
    """ Convert a JSON file to a CSV. Only properties (attributes) common to
    all features are retained. The arguments `fin` and `fout` may be either
    filenames or file-like objects.
    """

    if hasattr(fin, 'read'):
        jsdat = json.load(fin)
    else:
        with open(fin, 'r') as f:
            jsdat = json.load(f)

    # Load coordinates and properties
    X = []
    Y = []
    PROP = []
    for feat in jsdat['features']:
        x, y = feat['geometry']['coordinates']
        X.append(x)
        Y.append(y)
        PROP.append(feat['properties'])

    # Find subset of properties that exist in all features
    if len(PROP) > 1:
        s = set(PROP[0].keys())
        common_props = s.intersection(*[set(p.keys()) for p in PROP[1:]])
    else:
        common_props = PROP[0].keys()


    # Output
    outstr = 'x,y' + \
             reduce(lambda a,b: a+','+b, common_props) + \
             '\n'   # Construct header

    outstr += reduce(lambda a, b: a+'\n'+b,                         # Combine lines
                map(lambda x, y, p: str(x) + ',' + str(y) + ',' +   # Construct lines
                    reduce(lambda a, b: str(a) + ',' + str(b),      # Construct properties
                        [p[cp] for cp in common_props]),            # Filter spurious props
                    X, Y, PROP))                                    # Args for line lambda

    try:
        if hasattr(fout, 'read'):
            f = fout
        else:
            f = open(fout, 'w')
        f.write(outstr)

    finally:
        if not hasattr(fout, 'read'):
            f.close()

    return





"""
Geographical measurement and simple analysis module for Python 2.X.X. Provides
Point, Multipoint, Line, and Polygon classes, with methods for simple
measurements such as distance, area, and bearing.

Written by Nat Wilson (njwilson23@gmail.com)
"""

import math
import sys
import traceback
import numpy as np
import shapefile
from . import vtk
from . import geojson
from . import xyfile
from . import _shpfuncs
from .. import crs

from collections import deque
from .metadata import Metadata

try:
    from . import _cvectorgeo as _vecgeo
except ImportError:
    sys.stderr.write("falling back on slow _vectorgeo")
    from . import _vectorgeo as _vecgeo

try:
    import shapely.geometry as geometry
except ImportError:
    pass

try:
    import pyproj
    PYPROJ = True
    geod = pyproj.Geod(ellps="WGS84")
except ImportError:
    PYPROJ = False

class Geometry(object):
    """ This is the abstract base class for all geometry types, i.e. Point,
    Multipoints and subclasses thereof. """
    _geotype = None

    def __init__(self, crs=crs.CARTESIAN):
        self.properties = {}
        self._crs = crs
        return

    def _distance(self, pos0, pos1):
        """ Generic method for calculating distance between positions that
        respects CRS """
        if self._crs == crs.CARTESIAN:
            dist = _vecgeo.distance((pos0.x, pos0.y), (pos1.x, pos1.y))
        elif PYPROJ:
            _, _, dist = geod.inv(pos0.x, pos0.y, pos1.x, pos1.y, radians=False)
        else:
            dist = greatcircle(pos0, pos1)
        return dist

    def add_property(self, name, value):
        """ Insert a property (name -> value) into the properties dict, raising
        a NameError if the property already exists. Compared to directly
        manipulating the properties dict, this avoids accidents. """
        if name not in self.properties:
            self.properties[name] = value
        else:
            raise NameError("property '{0}' already exists in {1}".format(name, type(self)))
        return

    def delete_property(self, name):
        """ Equivalent to

            del self.properties[name]
        """
        if name in self.properties:
            del self.properties[name]
        return


class Point(Geometry):
    """ This defines the point class, from which x,y[,z] points can be
    constructed. """
    _geotype = "Point"

    def __init__(self, coords, data=None, properties=None, **kwargs):
        if properties is None: properties = {}
        if not hasattr(coords, "__iter__"):
            raise ValueError("input argument must be a sequence")
        super(Point, self).__init__(**kwargs)
        self.vertex = coords
        self._setxyz()
        self.data = Metadata(data, singleton=True)
        if hasattr(properties, "keys"):
            self.properties = properties
        return

    def __getitem__(self, idx):
        return self.vertex[idx]

    def __repr__(self):
        return 'Point(' + str(self.vertex) + ')'

    def __eq__(self, other):
        if hasattr(other, "vertex"):
            return (tuple(self.vertex) == tuple(other.vertex)) and \
                   (self.data == other.data) and \
                   (self.properties == other.properties)
        else:
            return False

    def _setxyz(self):
        """ Create convenience *x*, *y* (, *z*) attributes from vertices. """
        self.x = self.vertex[0]
        self.y = self.vertex[1]
        try:
            self.z = self.vertex[2]
            self.rank = 3
        except IndexError:
            self.z = None
            self.rank = 2
        return

    def isat(self, other, tol=0.001):
        """ Test whether point vertices are the same, regardless of properties or data.
        Tolarance is not geodetic. """
        return False not in [abs(a-b) <= tol for (a, b) in zip(self.vertices,
                                                               other.vertices)]

    def get_vertex(self):
        """ Return the Point vertex as a tuple. """
        return self.vertex

    def coordsxy(self, convert_to=False):
        """ Returns the x,y coordinates. Convert_to may be set to 'deg'
        or 'rad' for convenience.  """
        if convert_to == 'rad':
            return (self.x*3.14159/180., self.y*3.14159/180.)
        elif convert_to == 'deg':
            return (self.x/3.14159*180., self.y/3.14159*180.)
        else:
            return (self.x, self.y)

    def azimuth(self, other, spherical=False):
        """ Returns the compass azimuth from self to other in radians (i.e.
        clockwise, with north at 0). Returns NaN if points are coincident. """

        if self.coordsxy() == other.coordsxy():
            return np.nan

        if self._crs == crs.LONLAT and other._crs == crs.LONLAT:
            if not PYPROJ:
                raise CRSError("Azimuth on LONLAT grids requires pyproj")
            az1, _, _ = geod.inv(self.x, self.y, other.x, other.y)
            return az1 * math.pi / 180.0

        elif self._crs == crs.CARTESIAN and other._crs == crs.CARTESIAN:
            dx = other.x - self.x
            dy = other.y - self.y

            if dx > 0:
                if dy > 0:
                    return math.atan(dx/dy)
                elif dy < 0:
                    return math.pi - math.atan(-dx/dy)
                else:
                    return 0.5*math.pi
            elif dx < 0:
                if dy > 0:
                    return 2*math.pi - math.atan(-dx/dy)
                elif dy < 0:
                    return math.pi + math.atan(dx/dy)
                else:
                    return 1.5*math.pi
            else:
                if dy > 0:
                    return 0.0
                else:
                    return math.pi

        else:
            raise CRSError("Azimuth undefined for points in CRS {0} and "
                           "{1}".format(self._crs, other._crs))

    def walk(self, distance, bearing, azimuth=0.0, spherical=False):
        """ Wraps walk() """
        return walk(self, distance, bearing, azimuth=0.0, spherical=False)

    def distance(self, other):
        """ Returns a distance to another Point. If the coordinate system is
        geographical and a third (z) coordinate exists, it is assumed to have
        the same units as the real-world horizontal distance (i.e. meters). """
        if self._crs != other._crs:
            raise CRSError("Points must share the same coordinate system.")
        flat_dist = self._distance(self, other)
        if None in (self.z, other.z):
            return flat_dist
        else:
            return math.sqrt(flat_dist**2. + (self.z-other.z)**2.)

    def greatcircle(self, other):
        """ Return the great circle distance between two geographical points. """
        if not PYPROJ:
            raise CRSError("Great circle computations require pyproj")
        if not (self._crs == crs.LONLAT and other._crs == crs.LONLAT):
            raise CRSError("Great circle distances require both points to be "
                           "in geographical coordinates")
        az1, az2, dist = geod.inv(self.x, self.y, other.x, other.y, radians=False)
        return dist

    def shift(self, shift_vector):
        """ Shift point by the amount given by a vector. Operation occurs
        in-place """
        if len(shift_vector) != self.rank:
            raise GGeoError('Shift vector length must equal geometry rank.')

        self.vertex = tuple([a+b for a,b in zip(self.vertex, shift_vector)])
        self._setxyz()
        return self

    def as_geojson(self, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        writer = geojson.GeoJSONWriter(self, **kwargs)
        return writer.print_json()

    def to_geojson(self, f, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        try:
            if not hasattr(f, "write"):
                fobj = open(f, "w")
            else:
                fobj = f
            writer = geojson.GeoJSONWriter(self, **kwargs)
            writer.write_json(fobj)
        finally:
            if not hasattr(f, "write"):
                fobj.close()
        return writer

    def to_shapely(self):
        """ Returns a Shapely Point instance. """
        try:
            return geometry.Point(self.x, self.y, self.z)
        except NameError:
            raise ImportError('Shapely module did not import\n')


class MultipointBase(Geometry):
    """ Point cloud with associated attributes. This is a base class for the
    polyline and polygon classes. """
    _geotype = "MultipointBase"

    def __init__(self, vertices, data=None, properties=None, **kwargs):
        """ Create a feature with multiple vertices.

        vertices : a list of tuples containing point coordinates.

        data : is either `None` a list of point attributes, or a dictionary of
        point attributes. If `data` is not `None`, then it (or its values) must
        match `vertices` in length.
        """
        super(MultipointBase, self).__init__(**kwargs)
        vertices = list(vertices)
        if properties is None:
            properties = {}
        if len(vertices) > 0:
            self.rank = len(vertices[0])

            if self.rank > 3 or self.rank < 2:
                raise GInitError('Input must be doubles or triples\n')
            elif False in [self.rank == len(i) for i in vertices]:
                raise GInitError('Input must have consistent rank\n')
            else:
                self.vertices = [tuple(i) for i in vertices]

        else:
            self.rank = None
            self.vertices = []

        if hasattr(properties, "keys"):
            self.properties = properties

        self.data = Metadata(data)
        return

    def __repr__(self):
        if len(self) < 5:
            ppverts = str(self.vertices)
        else:
            ppverts = str(self.vertices[:2])[:-1] + "..." + str(self.vertices[-2:])[1:]
        return '{typ}({verts})>'.format(
                typ=str(type(self))[:-1], verts=ppverts, prop=self.properties)

    def __str__(self):
        if len(self) < 5:
            ppverts = str(self.vertices)
        else:
            ppverts = str(self.vertices[:2])[:-1] + "..." + str(self.vertices[-2:])[1:]
        return '{typ}\n{verts}\nProperties:{prop}'.format(
                typ=type(self), verts=ppverts, prop=self.properties)

    def __len__(self):
        return len(self.vertices)

    def __getitem__(self, key):
        if isinstance(key, (int, np.int64)):
            return Point(self.vertices[key], data=self.data[key],
                         properties=self.properties, crs=self._crs)
        elif isinstance(key, slice):
            return type(self)(self.vertices[key], data=self.data[key],
                              properties=self.properties, crs=self._crs)
        else:
            raise GGeoError('Index must be an integer or a slice object')

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise GGeoError('Indices must be integers')
        if getattr(value, "_geotype", None) == "Point":
            self.vertices[key] = value.vertex
            self.data[key] = list(value.data.values())[0]
        elif len(value) == self.rank:
            self.vertices[key] = value
            self.data[key] = None
        else:
            raise GGeoError('Cannot insert non-Pointlike value with '
                            'length != {0}'.format(self.rank))
        return

    def __delitem__(self, key):
        if len(self) > key:
            del self.vertices[key]
            del self.data[key]
        else:
            raise GGeoError('Index ({0}) exceeds length'
                            '({1})'.format(key, len(self)))
        return

    def __iter__(self):
        return (self[i] for i in range(len(self)))

    def __eq__(self, other):
        return hasattr(other, "_geotype") and \
               (self._geotype == other._geotype) and \
               (len(self) == len(other)) and \
               (False not in (a==b for a,b in zip(self, other)))

    def _bbox_overlap(self, other):
        """ Return whether bounding boxes between self and another geometry
        overlap.
        """
        reg0 = self.get_bbox()
        reg1 = other.get_bbox()
        return (reg0[0] < reg1[1] and reg0[1] > reg1[0] and
                reg0[2] < reg1[3] and reg0[3] > reg1[2])

    def get_bbox(self):
        """ Return the extents of a bounding box as
            (xmin, ymax, ymin, ymax, [zmin, zmin]).
        """
        if self.rank == 2:
            x, y = self.get_coordinate_lists()
            bbox = (min(x), max(x), min(y), max(y))
        elif self.rank == 3:
            x, y, z = self.get_coordinate_lists()
            bbox = (min(x), max(x), min(y), max(y), min(z), max(z))
        return bbox

    def print_vertices(self):
        """ Prints an enumerated list of indices. """
        for i, vertex in enumerate(self.vertices):
            print("{0}\t{1}".format(i, vertex))

    def get_vertices(self):
        """ Return vertices as a list of tuples. """
        return np.array(self.vertices)

    #def get_data(self, fields=None):
    #    """ Return data as an array, regardless of internal type. Optionally
    #    takes the keyword argument *fields*, which is an iterable listing the
    #    columns from the data dictionary to retrieve. """
    #    if hasattr(self.data, 'keys') and hasattr(self.data.values, '__call__'):
    #        if fields is not None:
    #            data = np.array([self.data[key] for key in fields])
    #        else:
    #            data = np.array(self.data.values())
    #    else:
    #        data = np.array(self.data)
    #    return data.T

    def get_coordinate_lists(self):
        """ Return X, Y, and Z lists. If self.rank == 2, Z will be
        zero-filled. """
        X = [i[0] for i in self.vertices]
        Y = [i[1] for i in self.vertices]
        if self.rank == 3:
            Z = [i[2] for i in self.vertices]
            return X, Y, Z
        else:
            return X, Y

    def shift(self, shift_vector):
        """ Shift feature by the amount given by a vector. Operation
        occurs in-place """
        if len(shift_vector) != self.rank:
            raise GGeoError('Shift vector length must equal geometry rank.')

        if self.rank == 2:
            f = lambda pt: (pt[0] + shift_vector[0], pt[1] + shift_vector[1])
        elif self.rank == 3:
            f = lambda pt: (pt[0] + shift_vector[0], pt[1] + shift_vector[1],
                            pt[2] + shift_vector[2])
        self.vertices = list(map(f, self.vertices))
        return self

    def _matmult(self, A, x):
        """ Return Ax=b """
        b = []
        for a in A:
            b.append(sum([ai * xi for ai, xi in zip(a, x)]))
        return b

    def rotate2d(self, thetad, origin=(0, 0)):
        """ Rotate rank 2 Multipoint around *origin* counter-clockwise by
        *thetad* degrees. """
        # First, shift by the origin
        self.shift([-a for a in origin])

        # Multiply by a rotation matrix
        theta = thetad / 180.0 * math.pi
        R = ((math.cos(theta), -math.sin(theta)),
             (math.sin(theta), math.cos(theta)))
        rvertices = [self._matmult(R, x) for x in self.vertices]
        self.vertices = rvertices

        # Shift back
        self.shift(origin)
        return self

    def _subset(self, idxs):
        """ Return a subset defined by index in *idxs*. """
        vertices = [self.vertices[i] for i in idxs]
        data = self.data.sub(idxs)
        subset = type(self)(vertices, data=data, properties=self.properties, crs=self._crs)
        return subset

    def flat_distances_to(self, pt):
        """ Return the "flat Earth" distance from each vertex to a point. """
        A = np.array(self.vertices)
        P = np.tile(np.array(pt.vertex), (A.shape[0], 1))
        d = np.sqrt(np.sum((A-P)**2, 1))
        return d

    def distances_to(self, pt):
        """ Return the distance from each vertex to a point. """
        d = [pt.distance(a) for a in self]
        return np.array(d)

    def greatcircles_to(self, pt):
        """ Return the great circle distances from each vertex to a point. """
        d = [pt.greatcircle(a) for a in self]
        return np.array(d)

    def nearest_point_to(self, pt):
        """ Returns the internal point that is nearest to pt (Point class).
        If two points are equidistant, only one will be returned.
        """
        distances = self.distances_to(pt)
        idx = np.argmin(distances)
        return self[idx]

    def get_extents(self):
        """ Calculate a bounding box. """
        def gen_minmax(G):
            """ Get the min/max from a single pass through a generator. """
            mn = mx = next(G)
            for x in G:
                mn = min(mn, x)
                mx = max(mx, x)
            return mn, mx
        # Get the min/max for a generator defined for each dimension
        return list(map(gen_minmax,
                    map(lambda i: (c[i] for c in self.vertices),
                        range(self.rank))))

    # This code should compute the convex hull of the points and then test the
    # hull's combination space
    # Alternatively, calculate the eigenvectors, rotate, and cherrypick the
    # points
    #def max_dimension(self):
    #    """ Return the two points in the Multipoint that are furthest
    #    from each other. """
    #    dist = lambda xy0, xy1: math.sqrt((xy1[0]-xy0[0])**2 +
    #                                      (xy1[1]-xy0[1])**2)

    #    P = [(p0, p1) for p0 in self.vertices for p1 in self.vertices]
    #    D = map(dist, (p[0] for p in P), (p[1] for p in P))
    #    return P[D.index(max(D))]

    def to_xyfile(self, fnm, fields=None, delimiter=' ', header=None):
        """ Write data to a delimited ASCII table.

        fnm         :   filename to write to

        kwargs:
        fields      :   specify the fields to be written (default all)

        Additional kwargs are passed to `xyfile.write_xy`.
        """
        xyfile.write_xy(self.get_vertices(), fnm, delimiter=delimiter, header=header)
        return

    def as_geojson(self, **kwargs):
        """ Print representation of internal data as a GeoJSON string.

        Parameters
        ----------
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        writer = geojson.GeoJSONWriter(self, crs=self._crs,
                    crs_fmt="ogc_crs_urn", **kwargs)
        return writer.print_json()

    def to_geojson(self, f, **kwargs):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : file-like object to recieve the GeoJSON string

        *kwargs* include:
        crs : coordinate reference system
        crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
        bbox : an optional bounding box tuple in the form (w,e,s,n)
        """
        try:
            if not hasattr(f, "write"):
                fobj = open(f, "w")
            else:
                fobj = f
            writer = geojson.GeoJSONWriter(self, crs=self._crs,
                        crs_fmt="ogc_crs_urn", **kwargs)
            writer.write_json(fobj)
        finally:
            if not hasattr(f, "write"):
                fobj.close()
        return writer

    def to_vtk(self, f, **kwargs):
        """ Write data to an ASCII VTK .vtp file. """
        if not hasattr(f, "write"):
            with open(f, "w") as fobj:
                vtk.mp2vtp(self, fobj, **kwargs)
        else:
            vtk.mp2vtp(self, f, **kwargs)
        return

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            _shpfuncs.write_multipoint2(self, fstem)
        elif self.rank == 3:
            _shpfuncs.write_multipoint3(self, fstem)
        else:
            raise IOError("rank must be 2 or 3 to write as a shapefile\n")
        return

class Multipoint(MultipointBase):
    """ Point cloud with associated attributes. This is a base class for the
    polyline and polygon classes. """
    _geotype = "Multipoint"

    def __init__(self, vertices, data=None, properties=None, **kwargs):
        """ Create a feature with multiple vertices.

        vertices : a list of tuples containing point coordinates.

        data : is either `None` a list of point attributes, or a dictionary of
        point attributes. If `data` is not `None`, then it (or its values) must
        match `vertices` in length.
        """
        if False not in (hasattr(v, "_geotype") and \
                         v._geotype == "Point" and \
                         v.rank == vertices[0].rank for v in vertices):
            points = vertices
            crs = points[0]._crs
            if False in (pt._crs == crs for pt in points):
                raise CRSError("All points must share the same CRS")

            keys = list(points[0].properties.keys())
            for pt in points[1:]:
                for key in keys:
                    if key not in pt.properties:
                        keys.pop(keys.index(key))

            ptdata = {}
            for key in keys:
                ptdata[key] = [pt.properties[key] for pt in points]

            if data is not None:
                ptdata.update(data)
            self.data = Metadata(ptdata)
            self.vertices = [pt.vertex for pt in points]
            self.rank = vertices[0].rank
            self.properties = properties
            self._crs = crs

        else:

            super(Multipoint, self).__init__(vertices,
                                             data=data, 
                                             properties=properties,
                                             **kwargs)
        return

    def within_radius(self, pt, radius):
        """ Return Multipoint of subset that is within *radius* of *pt*.
        """
        distances = self.distances_to(pt)
        indices = [i for i,d in enumerate(distances) if d <= radius]
        return self._subset(indices)

    def within_bbox(self, bbox):
        """ Return Multipoint subset that is within a square boundaing box
        given by (xmin, xmax, ymin, ymax). """
        filtbbox = lambda pt: (bbox[0] <= pt.vertex[0] <= bbox[1]) and \
                              (bbox[2] <= pt.vertex[1] <= bbox[3])
        indices = [i for (i, pt) in enumerate(self) if filtbbox(pt)]
        return self._subset(indices)


class ConnectedMultipoint(MultipointBase):
    """ Class for Multipoints in which vertices are assumed to be connected. """

    def length(self):
        """ Returns the length of the line/boundary. """
        points = [Point(i, crs=self._crs) for i in self.vertices]
        distances = [a.distance(b) for a, b in zip(points[:-1], points[1:])]
        return sum(distances)

    def segments(self):
        """ Returns an iterator of adjacent line segments. """
        return (self._subset((i,i+1)) for i in range(len(self)-1))

    def segment_tuples(self):
        """ Returns an iterator of adjacent line segments as coordinate tuples. """
        return ((self.vertices[i], self.vertices[i+1])
                for i in range(len(self.vertices)-1))

    def intersects(self, other):
        """ Return whether an intersection exists with another geometry. """
        interxbool = (_vecgeo.intersects(a[0][0], a[1][0], b[0][0], b[1][0],
                                         a[0][1], a[1][1], b[0][1], b[1][1])
                    for a in self.segments() for b in other.segments())
        if self._bbox_overlap(other) and (True in interxbool):
            return True
        else:
            return False

    def intersections(self, other):
        """ Return the intersections with another geometry. """
        interx = (_vecgeo.intersections(a[0][0], a[1][0], b[0][0], b[1][0],
                                        a[0][1], a[1][1], b[0][1], b[1][1])
                    for a in self.segments() for b in other.segments())
        return list(filter(lambda a: np.nan not in a, interx))

    def shortest_distance_to(self, pt):
        """ Return the shortest distance from a point on the Multipoint
        boundary to *pt* (Point) """
        point_dist = map(_vecgeo.pt_nearest,
                                [pt.vertex for seg in self.segments()],
                                [seg[0].vertex for seg in self.segments()],
                                [seg[1].vertex for seg in self.segments()])
        distances = [i[1] for i in point_dist]
        return min(distances)

    def nearest_on_boundary(self, pt):
        """ Returns the point on the Multipoint boundary that is nearest to pt
        (Point).

        Warning: If two points are equidistant, only one will be returned.
        """
        point_dist = list(map(_vecgeo.pt_nearest,
                                [pt.vertex for seg in self.segments()],
                                [seg[0].vertex for seg in self.segments()],
                                [seg[1].vertex for seg in self.segments()]))
        distances = [i[1] for i in point_dist]
        return Point(point_dist[distances.index(min(distances))][0],
                     properties=self.properties, crs=self._crs)

    def within_distance(self, pt, distance):
        """ Test whether a point is within *distance* of a ConnectedMultipoint. """
        #return distance <= self.shortest_distance_to(pt)
        return True in (distance >= seg.shortest_distance_to(pt)
                        for seg in self.segments())

class Line(ConnectedMultipoint):
    """ This defines the polyline class, from which geographic line
    objects can be constructed. Line objects consist of joined,
    georeferenced line segments.
    """
    _geotype = "Line"

    def add_vertex(self, vertex):
        """ Add a vertex to self.vertices. """
        if isinstance(vertex, Point):
            if self.rank == 2:
                self.vertices.append((vertex.x, vertex.y))
            elif self.rank == 3:
                self.vertices.append((vertex.x, vertex.y, vertex.z))
            self.data += vertex.data
        else:
            if self.rank == 2:
                self.vertices.append((vertex[0], vertex[1]))
            elif self.rank == 3:
                self.vertices.append((vertex[0], vertex[1], vertex[2]))
        return self

    def remove_vertex(self, index):
        """ Removes a vertex from the register by index. """
        pt = Point(self.vertices.pop(index), data=self.data[index])
        del self.data[index]
        return pt

    def extend(self, other):
        """ Combine two lines, provided that that the data formats are similar.
        """
        if self.rank == other.rank:
            if self._geotype == other._geotype:
                self.vertices.extend(other.vertices)
                self.data = self.data + other.data
            else:
                GGeoError('Cannot add inconsistent geometry types')
        else:
            GGeoError('Cannot add geometries with inconsistent rank')
        return self

    def cumlength(self):
        """ Returns the cumulative length by segment, prefixed by zero. """
        d = [0.0]
        pta = self[0]
        for ptb in self[1:]:
            d_ = pta.distance(ptb)
            d.append(d_ + d[-1])
            pta = ptb
        return d

    def displacement(self):
        """ Returns the distance between the first and last vertex. """
        return Point(self.vertices[0]).distance(Point(self.vertices[-1]))

    def direction(self):
        """ Returns a vector of the azimuth along a line at each point,
        weighted by neighbour position. """
        # The weights are calculated to give greater weight to azimuths to
        # nearby points
        # a' = w1 a1 + w2 a2
        #    = (a1 h2) / (h1 + h2) + (a2 h1) / (h1 + h2)
        #    = (a1 h2 + a2 h1) / (h1 + h2)
        azs = np.empty(len(self)-1)
        hs = np.empty(len(self)-1)
        pts = [pt for pt in self]
        for i in range(len(self)-1):
            azs[i] = pt[i].azimuth(pt[i+1])
            hs[i] = pt[i].distance(pt[i+1])
        weighted_azs = np.empty(len(self))
        weighted_azs[0] = azs[0]
        weighted_azs[-1] = azs[-1]
        for i in range(1, len(self)-1):
            raise NotImplementedError("need to phase unwrap")
            weighted_azs[i] = (azs[i]*hs[i+1] + azs[i+1]*hs[i]) / (hs[i] + hs[i+1])
        return weighted_azs

    def to_polygon(self):
        """ Returns a polygon. """
        return Polygon(self.vertices)

    def to_shapely(self):
        """ Returns a Shapely LineString instance. """
        try:
            if self.rank == 2:
                return geometry.LineString([(v[0], v[1]) for v in self.vertices])
            elif self.rank == 3:
                return geometry.LineString([(v[0], v[1], v[2]) for v in self.vertices])
        except NameError:
            raise GuppyError('Shapely module not available\n')

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            _shpfuncs.write_line2(self, fstem)
        elif self.rank == 3:
            _shpfuncs.write_line3(self, fstem)
        else:
            raise IOError("rank must be 2 or 3 to write as a shapefile\n")
        return

class Polygon(ConnectedMultipoint):
    """ This defines the polygon class, from which geographic
    polygons objects can be created. Polygon objects consist of
    point nodes enclosing an area.
    """
    _geotype = "Polygon"
    subs = []

    def __init__(self, vertices, data=None, properties=None, subs=None, **kwargs):
        ConnectedMultipoint.__init__(self, vertices, data=data, properties=properties, **kwargs)
        if vertices[0] != vertices[-1]:
            self.vertices.append(vertices[0])
        self.subs = subs if subs is not None else []
        return

    def __getitem__(self, key):
        if isinstance(key, slice):
            ind = key.indices(len(self))
            if len(self) != ((ind[1] - ind[0]) // ind[2]):
                return Line(self.vertices[key], data=self.data[key], 
                            properties=self.properties, crs=self._crs)
        return super(Polygon, self).__getitem__(key)

    def perimeter(self):
        """ Return the perimeter of the polygon. If there are sub-polygons,
        their perimeters are added recursively. """
        return self.length() + sum([p.perimeter() for p in self.subs])

    def area(self):
        """ Return the two-dimensional area of the polygon. If there are
        sub-polygons, their areas are subtracted. """
        a = 0.0
        for i in range(len(self.vertices)-1):
            a += 0.5 * abs((self.vertices[i][0] + self.vertices[i+1][0])
                         * (self.vertices[i][1] - self.vertices[i+1][1]))
        return a - sum(map(lambda p: p.area(), self.subs))

    def contains(self, pt):
        """ Returns True if pt is inside or on the boundary of the
        polygon, and False otherwise.
        """
        def possible(pt, v1, v2):
            """ Quickly assess potential for an intersection with an x+
            pointing ray. """
            x = pt.vertex[0]
            y = pt.vertex[1]
            if ( ((y > v1[1]) is not (y > v2[1]))
            and ((x < v1[0]) or (x < v2[0])) ):
                return True
            else:
                return False

        # Find how many boundaries a ray pointing out from point crosses
        bool2int = lambda tf: (tf and 1 or 0)
        rvertices = deque(self.vertices)
        rvertices.rotate(1)
        segments = [(v1, v2) for v1, v2 in zip(self.vertices, rvertices)
                    if possible(pt, v1, v2)]

        n_intersect = sum([bool2int(
            isinstance(ray_intersection(pt.vertex, seg[0], seg[1]), tuple))
            for seg in segments])

        if n_intersect % 2 == 1:    # If odd, point is inside so check subpolys
            if True not in (p.contains(pt) for p in self.subs):
                return True
        return False                # Point was outside or was in a subpoly

    def to_polyline(self):
        """ Returns a self-closing polyline. Discards sub-polygons. """
        return Line(self.vertices)

    def to_shapely(self):
        """ Returns a Shapely Polygon instance. """
        try:
            shp = geometry.Polygon(self.vertices,
                                   interiors=[p.vertices for p in self.subs])
        except NameError:
            raise ImportError('Shapely module did not import\n')
        return shp

    def to_shapefile(self, fstem):
        """ Save line to a shapefile """
        if self.rank == 2:
            _shpfuncs.write_poly2(self, fstem)
        elif self.rank == 3:
            _shpfuncs.write_poly3(self, fstem)
        else:
            raise IOError("rank must be 2 or 3 to write as a shapefile\n")
        return


class GuppyError(Exception):
    """ Base class for guppy module errors. """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message


class GInitError(GuppyError):
    """ Exception to raise when a guppy object fails to initialize. """
    def __init__(self, message=''):
        self.message = message


class GUnitError(GuppyError):
    """ Exception to raise there is a projected unit problem. """
    def __init__(self, message=''):
        self.message = message


class GGeoError(GuppyError):
    """ Exception to raise when a guppy object attempts an invalid transform. """
    def __init__(self, message=''):
        self.message = message

class CRSError(GuppyError):
    """ Exception to raise for invalid geodetic operations. """
    def __init__(self, message=''):
        self.message = message

def ray_intersection(pt, endpt1, endpt2, direction=0.0):
    """ Determines whether a ray intersects a line segment. If yes,
    returns the point of intersection. If no, return None. Input
    "points" should be tuples or similar. Input direction is in
    radians, and defines the ray path from pt.
    """
    m_ray = math.tan(direction)
    if endpt2[0] != endpt1[0]:
        m_lin = float(endpt2[1] - endpt1[1]) / float(endpt2[0] - endpt1[0])
        if m_ray == m_lin:      # Lines are parallel
            return
        else:
            x_int = ( (m_ray * pt[0] - m_lin * endpt1[0] - pt[1] + endpt1[1])
                / (m_ray - m_lin) )

        # Test that y_int is within segment and ray points toward segment
        if ( (x_int >= endpt1[0]) is not (x_int >= endpt2[0]) and
            (x_int-pt[0] > 0) is (math.cos(direction) > 0) ):
            y_int = ( (m_ray * m_lin * pt[0] - m_ray * m_lin * endpt1[0]
                + m_ray * endpt1[1] - m_lin * pt[1]) / (m_ray - m_lin) )
            return (x_int, y_int)
        else:
            return

    else:       # Line segment is vertical
        if direction % math.pi/2. == 0.0 and direction != 0.0:
            # Lines are parallel
            return
        x_int = float(endpt1[0])
        y_int = (x_int - pt[0]) * m_ray + pt[1]
        # Test that y_int is within segment and ray points toward segment
        if ( (y_int >= endpt1[1]) is not (y_int >= endpt2[1]) and
            (x_int-pt[0] > 0) is (math.cos(direction) > 0) ):
            return (x_int, y_int)
        else:
            return


def greatcircle(pta, ptb, method="vicenty"):
    """ Computes the great circle distance between n point pairs on a
    sphere. Returns a list of length (n-1)

    [pntlist] contains a list of point objects

    [method] may be "vicenty" (default) or "haversine". The Haversine
    method is roughly 20% faster, but may yield rounding errors when
    coordinates are antipodal.
    """

    radius = 6371.
    deg2rad = np.pi / 180.

    x1 = pta.x * deg2rad
    x2 = ptb.x * deg2rad
    y1 = pta.y * deg2rad
    y2 = ptb.y * deg2rad

    dx = x2 - x1
    dy = y2 - y1

    if method == "haversine":
        try:
            distance = 2 * radius * math.asin(math.sqrt((math.sin(dy /
                2.))**2 + math.cos(y1) * math.cos(y2) *
                (math.sin(dx / 2.))**2))
        except GGeoError:
            traceback.print_exc()

    elif method == "vicenty":
        try:
            a = math.sqrt((math.cos(y2) * math.sin(dx))**2 +
                (math.cos(y1) * math.sin(y2) - math.sin(y1) *
                math.cos(y2) * math.cos(dx))**2)
            b = (math.sin(y1) * math.sin(y2) + math.cos(y1) *
                math.cos(y2) * math.cos(dx))
            distance = radius * math.atan2(a, b)
        except ZeroDivisionError:
            raise GGeoError("Zero in denominator")
            return None
        except:
            traceback.print_exc()
    else:
        raise Exception("Distance method unrecognized")
        distance = np.nan

    return distance


def walk(start_pt, distance, bearing, azimuth=0.0, spherical=False):
    """ Returns the point reached when moving in a given direction for
    a given distance from a specified starting location.

        start_pt (point): starting location
        distance (float): distance to walk
        bearing (float): horizontal walk direction in radians
        azimuth (float): vertical walk direction in radians

        [NOT IMPLEMENTED]
        spherical (bool): use a spherical reference surface (globe)
    """
    if azimuth != 0.0:
        distxy = distance * math.sin(azimuth)
        dz = distance * math.cos(azimuth)
    else:
        distxy = distance
        dz = 0.0
    dx = distxy * math.sin(bearing)
    dy = distxy * math.cos(bearing)

    if start_pt.rank == 3:
        return Point((start_pt.x+dx, start_pt.y+dy, start_pt.z+dz))
    elif start_pt.rank == 2:
        if azimuth != 0:
            sys.stderr.write("Warning: start_pt has rank 2 but azimuth is "
                             "nonzero\n")
        return Point((start_pt.x+dx, start_pt.y+dy))


def sortby(A, B):
    """ Sort a list A by the values in an ordered list B. """
    if len(A) != len(B):
        raise GGeoError("A and B must be of the same length")
    comb = zip(B,A)
    comb.sort()
    return [i[1] for i in comb]


def points_to_multipoint(points):
    """ Merge *points* into a Multipoint instance. Point properties are stored
    as Multipoint data. All points must use the same CRS.
    """
    crs = points[0]._crs
    if False in (pt._crs == crs for pt in points):
        raise CRSError("All points must share the same CRS")

    keys = list(points[0].properties.keys())
    for pt in points[1:]:
        for key in keys:
            if key not in pt.properties:
                keys.pop(keys.index(key))

    ptdata = {}
    for key in keys:
        ptdata[key] = [pt.properties[key] for pt in points]

    vertices = [pt.vertex for pt in points]

    return Multipoint(vertices, data=ptdata, crs=crs)

def tighten(X, Z):
    """ Return a list of corrected measurements from observations of
    topography across a cross-section. The inputs are equal length
    lists of observed distance and elevation.

    Usage scenario: While surveying transects using a tape, the tape
    is anchored to the topographical surface, rather than directly
    between same-height endpoints.
    """
    if len(X) != len(Z):
        raise GGeoError('Observation vectors must have equal length')

    DZ = [z2-z1 for z2,z1 in zip(Z[1:], Z[:-1])]
    DX = [x2-x1 for x2,x1 in zip(X[1:], X[:-1])]

    DXt = [math.sqrt(x*x-z*z) for x,z in zip(DX, DZ)]

    Xt = map(lambda i: sum(DXt[:i]) + X[0], range(len(DX)))

    return zip(Xt, Z)





""" Convenience reader functions """

import os
import shapefile
from . import guppy
from . import geojson
from . import xyfile
from .. import crs

def _parsegeojsoncrs(crstup):
    """ From a tuple representing a GeoJSON (name,None) or (href,type) pair,
    return an appropriate karta crs instance. """
    if crstup[0] is None:       # use default as defined by spec
        return crs.LONLAT_WGS84
    elif crstup[1] is None:     # named CRS
        for c in crs.crslist:
            if c.urn == crstup[0]:
                return c
        return crs.CRS("unknown", "unknown", crstup[0])
    else:                       # linked CRS
        return crs.CRS("unknown", crstup[1], crstup[0])

def read_geojson(f):
    """ Read a GeoJSON object file and return a list of geometries """
    R = geojson.GeoJSONReader(f)
    geoms = R.iter_geometries()
    gplist = []
    for geom in geoms:
        coordsys = _parsegeojsoncrs(geom.crs)
        if isinstance(geom, geojson.Point):
            gplist.append(guppy.Point(geom.coordinates, crs=coordsys))
        elif isinstance(geom, geojson.MultiPoint):
            gplist.append(guppy.Multipoint(geom.coordinates, crs=coordsys))
        elif isinstance(geom, geojson.LineString):
            gplist.append(guppy.Line(geom.coordinates, crs=coordsys))
        elif isinstance(geom, geojson.Polygon):
            gplist.append(guppy.Polygon(geom.coordinates[0],
                                        subs=geom.coordinates[1:],
                                        crs=coordsys))
    return gplist

def read_geojson_features(f):
    """ Read a GeoJSON object file and return a list of features """
    R = geojson.GeoJSONReader(f)
    features = R.pull_features()
    print()
    geoms = []
    coordsys = _parsegeojsoncrs(R.getcrs())
    for (geom, properties, id) in features:

        if isinstance(geom, geojson.Point):
            (p, d) = _geojson_properties2guppy(properties, 1)
            geoms.append(guppy.Point(geom.coordinates, properties=p, data=d,
                                     crs=coordsys))

        elif isinstance(geom, geojson.MultiPoint):
            (p, d) = _geojson_properties2guppy(properties, len(geom.coordinates))
            geoms.append(guppy.Multipoint(geom.coordinates, properties=p, data=d,
                                          crs=coordsys))

        elif isinstance(geom, geojson.LineString):
            (p, d) = _geojson_properties2guppy(properties, len(geom.coordinates))
            geoms.append(guppy.Line(geom.coordinates, properties=p, data=d,
                                    crs=coordsys))

        elif isinstance(geom, geojson.Polygon):
            (p, d) = _geojson_properties2guppy(properties, len(geom.coordinates))
            geoms.append(guppy.Polygon(geom.coordinates[0], properties=p, data=d,
                                       subs=geom.coordinates[1:],
                                       crs=coordsys))
    return geoms

def _geojson_properties2guppy(properties, n):
    """ Takes a dictionary (derived from a GeoJSON properties object) and
    divides it into singleton properties and *n*-degree data. """
    props = {}
    data = {}
    for (key, value) in properties.items():
        if isinstance(value, list) or isinstance(value, tuple):
            if len(value) == n:
                data[key] = value
            else:
                raise ValueError("properties must be singleton or per-vertex")
        else:
            props[key] = value
    return props, data


def shape2point(shape):
    """ Convert a shapefile._Shape `shape` to a guppy.Point. """
    return guppy.Point(*shape.points)

def shape2line(shape):
    """ Convert a shapefile._Shape `shape` to a guppy.Line. """
    verts = shape.points
    return guppy.Line(verts)

def shape2poly(shape):
    """ Converts a shapefile._Shape `shape` to a guppy.Polygon. """
    verts = shape.points
    return guppy.Polygon(verts)

def shape2multipoint(shape):
    """ Converts a shapefile._Shape `shape` to a guppy.Polygon. """
    verts = shape.points
    return guppy.Multipoint(verts)

def get_filenames(stem, check=False):
    """ Given a filename basename, return the associated shapefile paths. If
    `check` is True, ensure that the files exist."""
    shp = stem + '.shp'
    shx = stem + '.shx'
    dbf = stem + '.dbf'
    if check:
        for fnm in (shp, shx, dbf):
            if not os.path.isfile(fnm):
                raise Exception('missing {0}'.format(fnm))
    return {'shp':shp, 'shx':shx, 'dbf':dbf}

def open_file_dict(fdict):
    """ Open each file in a dictionary of filenames and return a matching
    dictionary of the file objects. """
    files = {}
    for ext in fdict.keys():
        files[ext] = open(fdict[ext], 'r')
    return files

def read_shapefile(stem):
    """ Read a shapefile given `stem`, which is the name without an extension.
    """
    fnms = get_filenames(stem, check=True)

    try:
        files = open_file_dict(fnms)

        reader = shapefile.Reader(shp=files['shp'], shx=files['shx'],
                                  dbf=files['dbf'])
        features = []
        for shape in reader.shapes():
            if len(shape.points) > 0:
                if shape.shapeType == 1:
                    features.append(shape2point(shape))
                elif shape.shapeType == 3:
                    features.append(shape2line(shape))
                elif shape.shapeType == 5:
                    features.append(shape2poly(shape))
                elif shape.shapeType == 8:
                    features.append(shape2multipoint(shape))
                else:
                    raise NotImplementedError("cannot read shape type "
                                              "{0}".format(shape.shapeType))

    finally:
        for f in files.values():
            f.close()

    return features

def read_xyfile(f, delimiter='', header_rows=0, astype=guppy.Multipoint, coordrank=2):
    """ Read an ASCII delimited table and return a guppy object given by *astype*.
    """
    dat = xyfile.load_xy(f, delimiter=delimiter, header_rows=header_rows)
    ncols = dat.shape[1]
    if ncols >= coordrank:
        coords = dat[:,:coordrank]
        if ncols > coordrank:
            data = dat[:,coordrank:]
        else:
            data = None
        return astype(coords, data=data)
    else:
        raise IOError('data table has insufficient number of columns')




import os
from distutils.core import setup
from distutils.extension import Extension
import numpy

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
    include_dirs = [numpy.get_include()]
    ext = ".pyx"
except ImportError:
    USE_CYTHON = False
    include_dirs = []
    ext = ".c"
    print("Warning: Cython not imported")
    print("If C sources exist in the source tree, they will be used")

extensions = [Extension("karta.raster.cfill_sinks",
                        ["karta/raster/cfill_sinks"+ext],
                        include_dirs=include_dirs),
              Extension("karta.raster.crfuncs",
                        ["karta/raster/crfuncs"+ext],
                        include_dirs=include_dirs),
              Extension("karta.vector._cvectorgeo",
                        ["karta/vector/_cvectorgeo"+ext],
                        include_dirs=include_dirs)]

if USE_CYTHON:
    extensions = cythonize(extensions)

for extension in extensions:
    if False in (os.path.exists(src) for src in extension.sources):
        # C-extension sources missing, so don't try to build them
        extensions = []
        print("Warning: Extension source not found: {0}".format(extension.sources))
        print("Not building accelerated modules")
        break

setup(
    name = "karta",
    version = "0.2.2",
    author = "Nat Wilson",
    author_email = "njwilson23@gmail.com",
    packages = ["karta", "karta.vector", "karta.raster"],
    url = "http://njwilson23.github.io/karta/",
    description = "A tidy Python package for geospatial computation",
    long_description = """
Karta - tidy Python package for geospatial computation
======================================================

*Karta* is a Leatherman for geographic analyses. *Karta* provides an interface
for solving problems in Python/Python3 that works nicely with existing packages.
It provides simple and clean vector and raster data types, a small selection of
analysis functions, and the ability to read and write several useful formats.

Goals of *Karta* include providing a simple, lightweight, and fast set of tools
useful for everyday spatial analysis, as well as a flexible set of abstractions
upon which to build more advanced routines. *Karta* should be considered a work
in progress.

##CONTENTS AT A GLANCE

- vector
    - guppy:        Vector geometry classes (e.g. `Point`, `Multipoint`, `Line`, `Polygon`)
    - gpx:          GPX class for parsing and constructing GPX eXchange files
    - geojson:      Classes and functions for reading and writing GeoJSON
    - vtk:          XML-based VTK interface
    - shp\_funcs:   Shapefile-to-guppy conversions through _pyshp_ interface
    - stats:        Basic geostatistical functions
    - xyfile:       ASCII table functions

- raster
    - grid:         Basic Grid types, including `StructuredGrid` and `RegularGrid`
    - raster:       General purpose raster functions
    - aaigrid:      Grid subclass specifically for reading, writing, and manipulating ESRI ASCII grids
    - flow:         Stream flow functions
    - streamline:   Streamline calculation

- tests : unit tests


##FORMATS

*Karta* provides a basic working interface to several of common file formats.
Currently partially-supported are:

- vector
    - ASCII tables (XYZ) (r,w)
    - GeoJSON (r,w)
    - GPS eXchange (GPX) (r,w)
    - VTK (w)
    - ESRI Shapefiles via pyshp (r,w)
- raster
    - ESRI ASCII Grid (r,w)
    - USGS DEM (WIP)

## INSTALLATION

The easiest way to install is to use `pip`.

    cd karta/
    pip install .

## DEPENDENCIES

### Required

- Python 2.7+ or Python 3.2+
- numpy
- pyshp
- pyproj (required for geodetic calculations)

### Optional

- cython
- scipy

Cython is an optional dependency used to speed up select functions. In general,
enhanced-performance functions will then be called automatically when available,
otherwise *Karta* will fall back to numpy and pure-Python versions.

When installing from PyPI, C source code is provided and will be automatically
compiled if a suitable compiler is available.

## TESTING

To run all unit tests, execute

    python tests/runtests.py

##LICENSE

This software is provided under the MIT license.

###MIT License:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
""",
    download_url = "https://github.com/njwilson23/karta/archive/master.zip",
    classifiers = ["Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   "Development Status :: 3 - Alpha",
                   "Topic :: Scientific/Engineering"],
    license = "MIT License",
    ext_modules = extensions
)




""" This is a database of CRS IDs. """

from collections import namedtuple

CRS = namedtuple("CRS", ["name", "type", "urn"])

# List of the available coordinate systems
CARTESIAN = CRS("cartesian", "local", "urn:ogc:def:crs:EPSG::5806")
LONLAT_WGS84 = CRS("lonlat", "geographical", "urn:ogc:def:crs:EPSG::4326")

UPSNORTH_WGS84 = CRS("UPS North", "projected", "urn:ogc:def:crs:EPSG::32661")
UPSSOUTH_WGS84 = CRS("UPS South", "projected", "urn:ogc:def:crs:EPSG::32761")

NSIDCNORTH = CRS("NSIDC Sea Ice Sterographic North", "projected", "urn:ogc:def:crs:EPSG::3411")
NSIDCSOUTH = CRS("NSIDC Sea Ice Sterographic South", "projected", "urn:ogc:def:crs:EPSG::3412")

# Convenience aliases
LONLAT = LONLAT_WGS84       # Convenience



localnames = locals()
crslist = [c for c in localnames.values() if isinstance(c, CRS)]




""" Provides functions for reading and writing ESRI shapefiles and returning a
guppy object. """

import os
import datetime
import numbers
import shapefile
from shapefile import Reader, Writer

def property_field_type(value):
    """ Determine the appropriate dBase field type for *value* """
    if isinstance(value, numbers.Number):
        if isinstance(value, numbers.Integral):
            desc = "I"
        elif isinstance(value, numbers.Real):
            desc = "O"
        else:
            raise TypeError("cannot choose the correct dBase type for "
                            "{0}\n".format(type(value)))
    elif isinstance(value, str):
        desc = "C"
    elif isinstance(value, datetime.datetime):
        desc = "@"
    elif isinstance(value, datetime.date):
        desc = "D"
    elif isinstance(value, bool):
        desc = "L"
    else:
        raise TypeError("cannot choose the correct dBase type for "
                        "{0}\n".format(type(value)))
    return desc

def addfields(writer, properties):
    writer.field("ID", "I", "8")
    values = []
    for key in properties:
        value = properties[key]
        typ = property_field_type(value)
        length = "100"
        writer.field(key.upper(), typ, length)
        value.append(value)
    writer.record("0", *values)
    return

def addfields_points(writer, points):
    writer.field("ID", "I", "8")
    keys = list(points.data.keys())
    for key in keys:
        writer.field(key, property_field_type(points.data[key][0]), "100")
    for i, pt in enumerate(points):
        writer.record(str(i), *[pt.data[key] for key in keys])
    return

def write_multipoint2(mp, fstem):
    w = shapefile.Writer(shapeType=shapefile.POINT)
    for vertex in mp.vertices:
        w.point(*vertex)
    addfields_points(w, mp)
    w.save(fstem)
    return

def write_multipoint3(mp, fstem):
    w = shapefile.Writer(shapeType=shapefile.POINTZ)
    for vertex in mp.vertices:
        w.point(*vertex)
    addfields_points(w, mp)
    w.save(fstem)
    return

def write_line2(line, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYLINE)
    w.poly(shapeType=shapefile.POLYLINE, parts=[line.vertices])
    addfields(w, line.properties)
    w.save(fstem)
    return

def write_line3(line, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYLINEZ)
    w.poly(shapeType=shapefile.POLYLINEZ, parts=[line.vertices])
    addfields(w, line.properties)
    w.save(fstem)
    return

def write_poly2(poly, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYGON)
    w.poly(shapeType=shapefile.POLYGON, parts=[poly.vertices])
    addfields(w, poly.properties)
    w.save(fstem)
    return

def write_poly3(poly, fstem):
    w = shapefile.Writer(shapeType=shapefile.POLYGONZ)
    w.poly(shapeType=shapefile.POLYGONZ, parts=[poly.vertices])
    addfields(w, poly.properties)
    w.save(fstem)
    return

def write_shapefile(features, fstem):
    """ Write *features* to a shapefile. All features must be of the same type
    (i.e. Multipoints, Lines, Polygons). """
    if len(features) == 1:
        features[0].to_shapefile(fstem)
        return
    elif False in (f._geotype == features[0]._geotype for f in features[1:]):
        raise IOError("all features must be of the same type")
    elif False in (f.rank == features[0].rank for f in features[1:]):
        raise IOError("all features must have the same dimensionality")

    RankError = IOError("feature must be in two or three dimensions to write "
                        "as shapefile")
    if features[0]._geotype == "Multipoint":
        if features[0].rank == 2:
            typ = shapefile.POINT
        elif features[0].rank == 3:
            typ = shapefile.POINTZ
        else:
            raise RankError
    elif features[0]._geotype == "Line":
        if features[0].rank == 2:
            typ = shapefile.POLYLINE
        elif features[0].rank == 3:
            typ = shapefile.POLYLINEZ
        else:
            raise RankError
    elif features[0]._geotype == "Polygon":
        if features[0].rank == 2:
            typ = shapefile.POLYGON
        elif features[0].rank == 3:
            typ = shapefile.POLYGONZ
        else:
            raise RankError

    w = shapefile.Writer(shapeType=typ)
    if typ in (shapefile.POINT, shapefile.POINTZ):

        # add geometry
        for feature in features:
            for pt in feature:
                w.point(*pt.vertex)

        # add records
        w.field("ID", "I", "8")
        keys = set(features[0].data.keys())     # for testing similarity
        keylist = list(keys)                    # preserves order

        if len(keys) != 0 and \
           False not in (set(f.data.keys()) for f in features[1:]):
            for key in keylist:
                testvalue = features[0].data[key][0]
                w.field(key.upper(), property_field_type(testvalue), "100")
            i = 0
            for feature in features:
                for pt in feature:
                    w.record(str(i), *[pt.data[key] for key in keylist])
                    i += 1
        else:
            i = 0
            for feature in features:
                for pt in feature:
                    w.record(str(i))
                    i += 1

    else:

        # add geometry
        for feature in features:
            #print(feature)
            w.poly([feature.vertices])

        # add records
        w.field("ID", "I", "8")
        keys = set(features[0].properties.keys())   # for testing similarity
        keylist = list(keys)                        # preserves order

        if len(keys) != 0 and \
           False not in (set(f.data.keys()) for f in features[1:]):
            for key in features[0].properties:
                value = features[0].properties[key]
                typ = property_field_type(value)
                length = "100"
                w.field(key.upper(), typ, length)
            for (i, feature) in enumerate(features):
                values = []
                for key in keylist:
                    value.append(feature.properties[key])
                w.record(str(i), *values)
        else:
            for (i, feature) in enumerate(features):
                w.record(str(i))

    w.save(fstem)
    return

